%{
/* **************************************************************************** */
/*                                                                              */
/*  Copyright (c) 2011 University of Southern California (USC)                  */ 
/*                                                                              */
/*  All rights reserved.                                                        */
/*                                                                              */
/*  Permission is given to use, copy, and modify this software for any          */
/*  non-commercial purpose as long as this copyright notice is not removed      */
/*  All other uses, including redistribution in whole or in part, are           */
/*  forbidden without prior written permission.                                 */
/*                                                                              */
/*  This software is provided with absolutely no warranty and no support.       */
/*                                                                              */
/*  For further information contact: pedro@isi.edu                              */
/*                                                                              */
/* **************************************************************************** */

#define YYDEBUG 1

#include "structs.h"
#include "hashtable.h"

/* ------------------------------------------------------------------------------------ */

extern void 	exit(int);
extern int 	yylex();
       int 	yyerror(char* e);

extern symbol_t* getLabelRef(char *);
extern symbol_t* getSymbolRef(char *);
extern void 	 handleDataMapDirective(symbol_t *s, int dir, oper_t *arg);

extern int 	yacc_debug;			
extern int 	main_debug;			
extern int 	lineno;

/* ------------------------------------------------------------------------------------ */

int file_line = 0;
int program_line = 0;

extern int 	curr_pc;			/* used when simulating the execution	*/

extern int 	inst_cnt;			/* number of identifiers in the code 	*/
extern inst_t	program[MAX_PROG_SIZE];		/* the sequence of instructions 	*/
extern int	mem_addr;			
extern oper_t 	memory[MAX_MEMORY_SIZE];	/* where the values are stored 		*/

extern int 	 id_cnt;			/* number of identifiers in the code 	*/
extern symbol_t* ids[MAX_IDS];			/* table of identifiers 		*/
						/* each id has the address in the 	*/
						/* memory section of the program	*/

extern int 	 label_cnt;			/* number of labels in the code 	*/
extern symbol_t* labels[MAX_LABELS];		/* table of labels in the code  	*/
						/* each label has the address in the 	*/
						/* code section of the program		*/

extern hashtable* label_hash;
extern hashtable* id_hash;

/* ------------------------------------------------------------------------------------ */
/* Global Variables for helping the paerser						*/
/* ------------------------------------------------------------------------------------ */

inst_t    global_inst;
symbol_t* global_symbol;
oper_t*   global_oper;

int 	inside_data_section;
int 	inside_text_section;

%}

%union {
  int     i;
  double  d;
  char*   s;
  char    c;
  struct oper_t* op;
  struct symbol_t* sym;
  struct inst_t* ip;
}

%token SEC_DIR TEXT_DIR DATA_DIR WORD_DIR BYTE_DIR ASCII_DIR GLOBAL_DIR SKIP_DIR END_DIR
%token LPAREN RPAREN LBR RBR GOTO COMMA COLLON PUTPAR GETPAR RETURN CALL
%token PLUS MINUS TIMES DIV LOG_AND LOG_OR LOG_NOT BIT_AND BIT_OR BIT_NOT PHI
%token ATRIB ID IVAL DVAL CVAL SVAL NL REG_ID
%token LT GT LE GE EQ NEQ IF HALT NOP

%type <s>  ID
%type <i>  REG_ID
%type <i>  IVAL
%type <d>  DVAL
%type <c>  CVAL
%type <s>  SVAL
%type <i>  unary_oper bin_oper relop
%type <op>  oper 
%type <sym> label 
%type <ip> instr instr_ex

%start program
%%

program : section_list          	{  if(yacc_debug) printf(" > YACC: program -> section_list\n"); }
        ;

section_list : section section_list 	{ if(yacc_debug) printf(" > YACC: section_list -> section section_list\n"); }
        | section                   	{ if(yacc_debug) printf(" > YACC: section_list -> section\n"); }
        ;

section : text_section          	{ if(yacc_debug) printf(" > YACC: section -> text_section\n"); }
        | data_section          	{ if(yacc_debug) printf(" > YACC: section -> data_section\n"); }
	| empty_section			{ if(yacc_debug) printf(" > YACC: section -> empty_section\n"); }
        ;

empty_section : NL empty_section 	{ if(yacc_debug) printf(" > YACC: empty_section -> NL empty_section\n"); }
	| NL				{ if(yacc_debug) printf(" > YACC: empty_section -> NL\n"); }
	;

text_section : text_header code         { if(yacc_debug) printf(" > YACC: text_section -> section_header code\n"); }
        ;

data_section : data_header data_map_list { if(yacc_debug) printf(" > YACC: data_section -> data_header data_map_list\n"); }
        ;

data_map_list : data_map data_map_list 	{ if(yacc_debug) printf(" > YACC: data_map_list -> data_map data_map_list\n");	}
        |  data_map             	{ if(yacc_debug) printf(" > YACC: data_map_list -> data_map\n");		}
        ;

data_map : label WORD_DIR oper  NL 	{ 
  					  if(yacc_debug) { printf(" > YACC: data_map -> label .word oper NL\n"); }
  					  handleDataMapDirective($1, WORD_DIR, $3);
					}
	| label BYTE_DIR oper  NL 	{ 
  					  if(yacc_debug) { printf(" > YACC: data_map -> label .byte oper NL\n"); }
  					  handleDataMapDirective($1, BYTE_DIR, $3);
					}
	| label ASCII_DIR oper NL 	{ 
  					  if(yacc_debug) { printf(" > YACC: data_map -> label .ascii oper NL\n"); }
  					  handleDataMapDirective($1, ASCII_DIR, $3);
					}
	| label SKIP_DIR oper NL	{ 
  					  if(yacc_debug) { printf(" > YACC: data_map -> label .skip oper NL\n"); }
  					  handleDataMapDirective($1, SKIP_DIR, $3);
					}
	| NL				{ 
					  if(yacc_debug) printf(" > YACC: data_map -> NL\n"); 
					}
        ;

data_header : SEC_DIR DATA_DIR  NL 	{ 
					  if(yacc_debug) printf(" > YACC: data_header -> .section .data\n");  
                                          inside_text_section = 0;
                                          inside_data_section = 1;
					  program_line = 0;
					}
	|  SEC_DIR error NL		{
					  if(yacc_debug) printf(" > YACC: data_header -> .section error\n");  
					  printf(" *** Parsing error: Expecting a \'.data\' token\n");
                                          inside_text_section = 0;
                                          inside_data_section = 1;
					  program_line = 0;
					}
        ;

text_header : SEC_DIR TEXT_DIR  header_flags NL { 
					if(yacc_debug) printf(" > YACC: text_header -> .section .text flags\n"); 
                                        inside_text_section = 1;
                                        inside_data_section = 0;
					program_line = 0;
                                }
	|	SEC_DIR TEXT_DIR error NL	{
					  if(yacc_debug) printf(" > YACC: text_header -> .section error\n"); 
					  printf(" *** Parsing error: Something wrong in the text header declaration.\n");
                                          inside_text_section = 1;
                                          inside_data_section = 0;
					  program_line = 0;
				}
        ;

header_flags : ',' SVAL          { if(yacc_debug) printf(" > YACC: header_flags -> ',' \"string\" \n"); }
        |       SVAL             { if(yacc_debug) printf(" > YACC: header_flags -> \"string\" \n"); }
        |                       { if(yacc_debug) printf(" > YACC: header_flags -> \n"); }
        ;

code	: inst code 		{  if(yacc_debug) printf(" > YACC: inst code (1)\n"); }
	| inst 			{  if(yacc_debug) printf(" > YACC: inst code (2)\n"); }
	;

inst 	: label instr_ex 	{ if(yacc_debug) printf(" > YACC: label instr_ex\n"); yyerrok; }
	| instr_ex  		{ if(yacc_debug) printf(" > YACC: instr_ex\n"); yyerrok; }
	| error			{ if(yacc_debug) printf(" > YACC: error in line %d\n",lineno); 
				  printf(" *** Parsing error: Something wrong in the the instruction in line %d,\n",file_line);
					  yyerrok;
				}
	; 

instr_ex : instr  { 
		  program[curr_pc].op1 = global_inst.op1; 
                  program[curr_pc].op2 = global_inst.op2; 
                  program[curr_pc].op3 = global_inst.op3;
                  program[curr_pc].operator = global_inst.operator;
                  program[curr_pc].type = global_inst.type;
                  program[curr_pc].line = global_inst.line;
                  curr_pc++;
		  inst_cnt++;
		  clearInstruction(&global_inst);
		  if(yacc_debug) printf(" > YACC: instr\n");
		  program_line++;
		}
	;

instr : oper ATRIB oper bin_oper oper NL {
          global_inst.type = T_OPER;
          copyOperand(&global_inst.op1, (oper_t*)$1);
          copyOperand(&global_inst.op2, (oper_t*)$3);
          copyOperand(&global_inst.op3, (oper_t*)$5);
          global_inst.line = program_line;
        }
        | oper ATRIB unary_oper oper NL  { 
          global_inst.type = T_UNARY;
          copyOperand(&global_inst.op1, (oper_t*)$1);
          copyOperand(&global_inst.op2, (oper_t*)$4);
          global_inst.line = program_line;
        }
        | oper ATRIB oper NL {
          global_inst.type = T_COPY;
          copyOperand(&global_inst.op1, (oper_t*)$1);
          copyOperand(&global_inst.op2, (oper_t*)$3);
          global_inst.line = program_line;
        } 
        | array_elem ATRIB oper NL {
          global_inst.type = T_MSTORE;
          copyOperand(&global_inst.op3, (oper_t*)$3);
          global_inst.line = program_line;
        }
        | oper ATRIB array_elem NL {
          global_inst.type = T_MLOAD;
          copyOperand(&global_inst.op3, (oper_t*)$1);
          global_inst.line = program_line;
        } 
        | oper ATRIB BIT_AND oper NL {
          global_inst.type = T_ADDRESSOF;
          copyOperand(&global_inst.op1, (oper_t*)$1);
          copyOperand(&global_inst.op2, (oper_t*)$4);
          global_inst.line = program_line;
	}
        | oper ATRIB TIMES oper NL {
          global_inst.type = T_DEREF;
          copyOperand(&global_inst.op1, (oper_t*)$1);
          copyOperand(&global_inst.op2, (oper_t*)$4);
          global_inst.line = program_line;
	}
        | TIMES oper ATRIB oper NL {
          global_inst.type = T_PTRASSIGN;
          copyOperand(&global_inst.op1, (oper_t*)$2);
          copyOperand(&global_inst.op2, (oper_t*)$4);
          global_inst.line = program_line;
	}
	| IF oper GOTO oper NL {
          global_inst.type = T_BOOL_COND;
          copyOperand(&global_inst.op1, (oper_t*)$2);
          copyOperand(&global_inst.op2, (oper_t*)$4);
          global_inst.line = program_line;
        }
        | IF oper relop oper GOTO oper NL {
          global_inst.type = T_COND;
          copyOperand(&global_inst.op1, (oper_t*)$2);
          copyOperand(&global_inst.op2, (oper_t*)$4);
          copyOperand(&global_inst.op3, (oper_t*)$6);
          global_inst.line = program_line;
        }
	| IF LPAREN oper RPAREN GOTO oper NL {
          global_inst.type = T_BOOL_COND;
          copyOperand(&global_inst.op1, (oper_t*)$3);
          copyOperand(&global_inst.op2, (oper_t*)$6);
          global_inst.line = program_line;
        }
        | IF LPAREN oper relop oper RPAREN GOTO oper NL {
          global_inst.type = T_COND;
          copyOperand(&global_inst.op1, (oper_t*)$3);
          copyOperand(&global_inst.op2, (oper_t*)$5);
          copyOperand(&global_inst.op3, (oper_t*)$8);
          global_inst.line = program_line;
        }
        | IF error GOTO oper NL {
          global_inst.type = T_GOTO;
          copyOperand(&global_inst.op1, (oper_t*)$4);
	  printf(" *** Warn: Invalid predicate in \'if\' instruction in line %d - unconditional GOTO inserted\n",
	   lineno);
          global_inst.line = program_line;
	  yyerrok;
        }
        | GOTO oper NL {
          global_inst.type = T_GOTO;
          copyOperand(&global_inst.op1, (oper_t*)$2);
          global_inst.line = program_line;
        }
        | GOTO NL {
          global_inst.type = T_NOP;
	  printf(" *** Warn: Empty GOTO target in line %d - NOP inserted\n",lineno);
          global_inst.line = program_line;
        }
        | GOTO error NL {
          global_inst.type = T_NOP;
	  printf(" *** Warn: Incorrect GOTO instruction in line %d - NOP inserted\n",lineno);
          global_inst.line = program_line;
	  yyerrok;
        }
        | HALT NL {
          global_inst.type = T_HALT;
          global_inst.line = program_line;
        }
        | NOP NL {
          global_inst.type = T_NOP;
          global_inst.line = program_line;
        }
        | PUTPAR oper NL {
          global_inst.type = T_PUTPAR;
          copyOperand(&global_inst.op1, (oper_t*)$2);
          global_inst.line = program_line;
        }
        | PUTPAR error NL {
          global_inst.type = T_NOP;
          global_inst.op1.type = O_NULL;
	  printf(" *** Warn: Invalid \'putparam\' instruction in line %d - NOP inserted\n",lineno);
          global_inst.line = program_line;
	  yyerrok;
        }
        | GETPAR oper NL {
          global_inst.type = T_GETPAR;
          copyOperand(&global_inst.op1, (oper_t*)$2);
          global_inst.line = program_line;
        }
        | GETPAR error NL {
          global_inst.type = T_NOP;
          global_inst.op1.type = O_NULL;
	  printf(" *** Warn: Invalid \'getparam\' instruction in line %d - NOP inserted\n",lineno);
          global_inst.line = program_line;
	  yyerrok;
        }

        | oper ATRIB CALL oper COMMA oper NL {
          global_inst.type = T_CALL;
          copyOperand(&global_inst.op1, (oper_t*)$1);
          copyOperand(&global_inst.op2, (oper_t*)$4);
          copyOperand(&global_inst.op3, (oper_t*)$6);
          global_inst.line = program_line;
        }
        | oper ATRIB CALL oper NL {
          global_inst.type = T_CALL;
          copyOperand(&global_inst.op1, (oper_t*)$1);
          copyOperand(&global_inst.op2, (oper_t*)$4);
          global_inst.op3.type = O_NULL;
          global_inst.line = program_line;
        }
        | oper ATRIB CALL error NL {
          global_inst.type = T_NOP;
          global_inst.op1.type = O_NULL;
	  printf(" *** Warn: Incorrect CALL instruction in line %d - NOP inserted\n",lineno);
          global_inst.line = program_line;
	  yyerrok;
        }
	| CALL oper COMMA oper NL {
          global_inst.type = T_CALL;
          global_inst.op1.type = O_NULL;
          copyOperand(&global_inst.op2, (oper_t*)$2);
          copyOperand(&global_inst.op3, (oper_t*)$4);
          global_inst.line = program_line;
	}
	| CALL error NL {
          global_inst.type = T_NOP;
          global_inst.op1.type = O_NULL;
	  printf(" *** Warn: Incorrect CALL instruction in line %d - NOP inserted\n",lineno);
          global_inst.line = program_line;
	  yyerrok;
	}
        | RETURN oper NL {
          global_inst.type = T_RETURN;
          copyOperand(&global_inst.op1, (oper_t*)$2);
          global_inst.line = program_line;
        }
	| RETURN NL {
          global_inst.type = T_RETURN;
          global_inst.op1.type = O_NULL;
          global_inst.line = program_line;
	}
	| RETURN error NL {
          global_inst.type = T_RETURN;
          global_inst.op1.type = O_NULL;
	  printf(" *** Warn: Incorrect RETURN instruction in line %d - NOP inserted\n",lineno);
          global_inst.line = program_line;
	  yyerrok;
	}
        | NL 	{
          global_inst.type = T_NOP;
          global_inst.op1.type = O_NULL;
          global_inst.line = program_line;
	}
	;

label : ID COLLON {
              	  int found; 
		  symbol_t *sym;

		  if(inside_text_section){ 
		    found = 0;
  		    sym = lookupItem(label_hash, &found, $1);
		    if(sym != NULL){
		      printf(" *** Error: Label %s in line %d multiply defined. Previous definition ignored.\n",$1,file_line);
		    }   
              	    sym = getLabelRef($1);
		    global_symbol = sym;
                    sym->address = program_line;
		    if(yacc_debug){
		      printf(" > YACC: Setting label %s to program counter %d\n",$1,program_line);
 		    }
		    $$ = (struct symbol_t*)sym;
		  } /* --- end of inside text section --- */

		  if(inside_data_section){ 
		    found = 0;
  		    sym = lookupItem(id_hash, &found, $1);
		    if(sym != NULL){
		      printf(" *** Error: Variable %s in line %d multiply defined. Previous definition ignored.\n",$1,file_line);
		    }   
              	    sym = getSymbolRef($1);
		    global_symbol = sym;
  		    sym->address = -1;
		    if(yacc_debug){
		      printf(" > YACC: Mapping variable %s to address %d and size %d\n",sym->tag, sym->address, sym->size);
 		    }
		    $$ = (struct symbol_t*)sym;
		    sym = NULL;
		  } /* --- end of inside data section --- */
        	}
	;

oper 	: ID  	{ oper_t *op;
		  if(yacc_debug){
		    printf(" > YACC: Debug ID = %s\n",$1);
		  }
		  op = createStringValOperand($1);
		  $$ = (struct oper_t*)op;
		  global_oper = op;
		}
	| IVAL  { oper_t *op;
		  if(yacc_debug){
		    printf(" > YACC: Debug IVAL = %d\n",$1);
		  }
		  op = createIntegerValOperand($1);
		  $$ = (struct oper_t*)op;
		  global_oper = op;
		}
	| DVAL  { oper_t *op;
		  if(yacc_debug){
		    printf(" > YACC: Debug DVAL = %f\n",$1);
		  }
		  op = createDoubleValOperand($1);
		  $$ = (struct oper_t*)op;
		  global_oper = op;
		}
	| CVAL  { oper_t *op;
		  if(yacc_debug){
		    printf(" > YACC: Debug CVAL = %c\n",$1);
		  }
		  op = createCharValOperand($1);
		  $$ = (struct oper_t*)op;
		  global_oper = op;
		}
	| SVAL  { oper_t *op;
		  if(yacc_debug){
		    printf(" > YACC: Debug SVAL = \"%s\"\n",$1);
		  }
		  op = createStringValOperand($1);
		  $$ = (struct oper_t*)op;
		  global_oper = op;
		}
	| REG_ID { oper_t *op;
		  if(yacc_debug){
		    if($1 < 0){
		      if($1 == PC_REG) printf(" > YACC: Debug $pc Register\n");
		      if($1 == SP_REG) printf(" > YACC: Debug $sp Register\n");
		      if($1 == FP_REG) printf(" > YACC: Debug $fp Register\n");
		    } else {
		      printf(" > YACC: Debug Register = $r%d\n",$1);
		    }
		  }
		  op = createRegisterValOperand($1);
		  $$ = (struct oper_t*)op;
		  global_oper = op;
		}
	;

array_elem : ID LBR oper RBR {
	        oper_t *tmp;
		if(yacc_debug){
		  printf(" > YACC: Debug array_elem : ID [ oper ]\n");
		}
	        tmp = (oper_t*)createStringValOperand($1);
                copyOperand(&global_inst.op1, tmp);
                copyOperand(&global_inst.op2, (oper_t*)$3);
             } 
	; 

bin_oper : PLUS 	{ global_inst.operator = PLUS;	} 
         | MINUS 	{ global_inst.operator = MINUS;	} 
         | TIMES 	{ global_inst.operator = TIMES;	} 
         | DIV 		{ global_inst.operator = DIV;	}	 
         | BIT_AND 	{ global_inst.operator = BIT_AND;	} 
         | BIT_OR 	{ global_inst.operator = BIT_OR;	} 
         | LOG_AND 	{ global_inst.operator = LOG_AND;	} 
         | LOG_OR 	{ global_inst.operator = LOG_OR;	} 
         | PHI 		{ global_inst.operator = PHI;	} 
	;

relop 	: LT 		{ global_inst.operator = LT;	} 
      	| GT 		{ global_inst.operator = GT;	} 
      	| LE 		{ global_inst.operator = LE;	} 
      	| GE 		{ global_inst.operator = GE;	} 
      	| EQ 		{ global_inst.operator = EQ;	}
      	| NEQ 		{ global_inst.operator = NEQ;	}
	;

unary_oper : LOG_NOT 	{ global_inst.operator = LOG_NOT;	}
 	| BIT_NOT 	{ global_inst.operator = BIT_NOT;	}
      	| MINUS 	{ global_inst.operator = MINUS;	}
	; 

%%

#include <stdio.h>

/* ----------------------------------------------------------------------- */
int yyerror(char *s) {

  /* 
  fflush(stdout);
  fprintf(stderr, " *** Error(s) encountered while processing input file in line %d\n",lineno+1);
  */
}

/* ----------------------------------------------------------------------- */
/* Assumes valid hex characters */
void hextodec(char *resul, char *orig) {
   int x = 0;
   /* hex string ends with h/H */
   while((*orig != 'h') && (*orig != 'H')){
      x = x << 4;
      *orig |= 32;                /* convert to lower case */
      if(*orig >= 'a')
          x |= *orig -'a' + 10;
      else
          x |= *orig - '0';
      orig++;
   }
   sprintf(resul, "%d", x);
}

/* ----------------------------------------------------------------------- */
/* --- Assumes valid oct characters --- */
void octtodec(char *resul, char *orig) {
   int x = 0;
   /* oct string ends with o/O */
   while((*orig != 'o') && (*orig != 'O')){
      x = x << 3;
      x |= *orig - '0';
      orig++;
   }
   sprintf(resul, "%d", x);
}

/* ----------------------------------------------------------------------- */
/* --- Assumes valid bin characters --- */
void bintodec(char *resul, char *orig) {
   int x = 0;
   /* bin string ends with b/B */
   while((*orig != 'b') && (*orig != 'B')){
      x = x << 1;
      x |= *orig - '0';
      orig++;
   }
   sprintf(resul, "%d", x);
}

/* ----------------------------------------------------------------------- */
void handleDataMapDirective(symbol_t *s, int dir, oper_t *arg){
  oper_t *op; 
  symbol_t *sym;

  sym = global_symbol;
  op = global_oper;
  switch(dir){
    case WORD_DIR:
      if(op->type == O_INTEGER){ 
        sym->size = 1;
        sym->address = mem_addr;
        mem_addr += 1;
        sym->initvalue = arg;
      } else {
        printf(" *** Error: Invalid (non-integer) value for .word directive of variable %s\n",sym->tag);
      }
      break;
    case BYTE_DIR:
      if(op->type == O_CHAR){ 
        sym->size = 1;
        sym->address = mem_addr;
        mem_addr += 1;
        sym->initvalue = arg;
      } else {
        printf(" *** Error: Invalid (non-char) value for .byte directive of variable %s\n",sym->tag);
      }
      break;
    case SKIP_DIR:
      if(op->type == O_INTEGER){
        sym->size = op->u.ival;
        sym->address = mem_addr;
        mem_addr += op->u.ival;
        sym->initvalue = NULL;
      } else {
        printf(" *** Error: Invalid (non-integer) value for .skip directive of variable %s\n", sym->tag);
      }
      break;
    case ASCII_DIR:
      if(op->type == O_STRING){
        sym->size = strlen(op->u.sval);
        sym->address = mem_addr;
        mem_addr += strlen(op->u.sval)+1;
        sym->initvalue = arg;
      } else {
        printf(" *** Error: Invalid (non-string) value for .ascii directive of variable %s\n", 
	  sym->tag);
      }
      break;
    default:
      break;
  } /* end of switch */

/*
  if(yacc_debug){
    printf(" > YACC: setting the initial value of %s at mem address %d to ",sym->tag,sym->address);
    printOperand(sym->initvalue);
    printf("\n");
  }
*/

}
/* ----------------------------------------------------------------------- */

