/* **************************************************************************** */
/*                                                                              */
/*  Copyright (c) 2011 University of Southern California (USC)                  */ 
/*                                                                              */
/*  All rights reserved.                                                        */
/*                                                                              */
/*  Permission is given to use, copy, and modify this software for any          */
/*  non-commercial purpose as long as this copyright notice is not removed      */
/*  All other uses, including redistribution in whole or in part, are           */
/*  forbidden without prior written permission.                                 */
/*                                                                              */
/*  This software is provided with absolutely no warranty and no support.       */
/*                                                                              */
/*  For further information contact: pedro@isi.edu                              */
/*                                                                              */
/* **************************************************************************** */

/* -------------------------------------------------------------------- */
/*									*/
/* -------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

/* -------------------------------------------------------------------- */
#include "structs.h"
#include "hashtable.h"

/* -------------------------------------------------------------------- */
int hash(hashtable *ht, char *str) { 
  int i, n, k = 0;
  char c;

  if(str == NULL)
    return 0;
  n = strlen(str);
  for(i = 0; i < n; i++){
    c = str[i];
    k += (int)c;
  }
  return (k % ht->size); 
}

/* -------------------------------------------------------------------- */
hashbucket* newEntry(int k, symbol_t *sym){
  hashbucket* res;

  res = (hashbucket*)malloc(sizeof(hashbucket));
  if(res == NULL){
   return NULL;
  }
  res->key = k;
  res->sym = sym;
  if(res->sym == NULL){
    free(res);
   return NULL;
  }
  return res;
}

/* -------------------------------------------------------------------- */
void deleteEntry(hashbucket *b){
  if(b != NULL){
    if(b->sym != NULL)
      //destroySymbol(b->sym);
      b->sym = NULL;
  }
  b->key = 0;
  b->next = NULL;
  free(b);
}

/* -------------------------------------------------------------------- */
hashtable* createHashTable(int size){
  int i;
  hashtable* res;

  if(size <= 0)
    size = hashtable_default_size;

  res = (hashtable*)malloc(sizeof(hashtable));
  res->size = size;
  res->table = (hashbucket**)malloc(sizeof(hashbucket*)*size);
  for(i = 0; i < size; i++){
    res->table[i] = NULL;
  }
  return res;
}

/* -------------------------------------------------------------------- */
void destroyHashTable(hashtable *ht){
  int i;
  hashbucket *b, *aux;

  if(ht == NULL)
    return;
  for(i = 0; i < ht->size; i++){
    b = ht->table[i];
    while(b != NULL){
      aux = b->next;
      b->next = NULL;
      deleteEntry(b);
      b = aux;
    }
    ht->table[i] = NULL;
  }
  free(ht->table);
  free(ht);
}

/* -------------------------------------------------------------------- */
hashbucket* getBucketList(hashtable *ht, char *it) {
  int k;
  hashbucket *b;
  k = hash(ht, it);
  b = ht->table[k];
  return b;
}

/* -------------------------------------------------------------------- */
symbol_t* lookupItem(hashtable *ht, int *found, char *it) {
  hashbucket *b;

  b = getBucketList(ht,it);
  while(b != NULL){
    if(b->sym != NULL){
      if(strcmp(it, b->sym->tag) == 0){
        *found = 1;
        return b->sym;
      }
    }
    b = b->next;
  }
  *found = 0;
  return NULL;
}

/* -------------------------------------------------------------------- */
symbol_t* insertItem(hashtable *ht, int idx, symbol_t *sym){
  int i, f;
  symbol_t* s;
  hashbucket *b, *aux;

  if(sym == NULL)
    return NULL;
  f = 0;
  s = lookupItem(ht, &f, sym->tag);
  if(f != 0){
    return s;
  } else {
    i = hash(ht,sym->tag);
    b = ht->table[i];
    aux = newEntry(idx,sym);
    aux->next = ht->table[i];
    ht->table[i] = aux;
    return sym;
  }
}

/* -------------------------------------------------------------------- */
void removeItem(hashtable *ht, char *it){
  int k, f;
  symbol_t* s;
  hashbucket *b, *aux;

  f = 0;
  s = lookupItem(ht, &f, it);
  if(f == 0){
    return;
  } else {
    k = hash(ht,it);
    b = ht->table[k];
    if(b == NULL)
      return;
    if(strcmp(b->sym->tag,it) == 0){
      ht->table[k] = b->next;
      b->next = NULL;
      deleteEntry(b);
      return;
    }
    while(b->next->next != NULL){
      if(strcmp(b->next->sym->tag,it) == 0){
	aux = b->next;
	b->next = aux->next;
	aux->next = NULL;
	deleteEntry(aux);
        return;
      }
      b = b->next;
    }
  }
}

