/* **************************************************************************** */
/*                                                                              */
/*  Copyright (c) 2011 University of Southern California (USC)                  */ 
/*                                                                              */
/*  All rights reserved.                                                        */
/*                                                                              */
/*  Permission is given to use, copy, and modif/y this software for any          */
/*  non-commercial purpose as long as this copyright notice is not removed      */
/*  All other uses, including redistribution in whole or in part, are           */
/*  forbidden without prior written permission.                                 */
/*                                                                              */
/*  This software is provided with absolutely no warranty and no support.       */
/*                                                                              */
/*  For further information contact: pedro@isi.edu                              */
/*                                                                              */
/* **************************************************************************** */

/* ---------------------------------------------------------------------------- */
/* $Id: main.c,v 1.3 2010/11/01 12:26:38 pedro Exp $ 				*/
/* ---------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

/* --------------------------------------------------------------------------- */

#include "structs.h"
#include "hashtable.h"

/* ---------------------------------------------------------------------------- */

extern symbol_t* getLabelRef(char *st);
extern symbol_t* getFunctionRef(char *st);
extern symbol_t* getSymbolRef(char *st);

/* ---------------------------------------------------------------------------- */

int		version;
void 		simulateV5();
void 		simulateV6();
extern int 	yyparse();
void 		io_proc(inst_t *ins, symbol_t *sym);

/* ---------------------------------------------------------------------------- */
/* Global Variables 								*/
/* ---------------------------------------------------------------------------- */
hashtable* label_hash;
hashtable* id_hash;
hashtable* fct_hash; 

FILE *yyin;

/* ---------------------------------------------------------------------------- */
int 	 inst_cnt;			/* the index into the program table.	*/
int 	 curr_pc;			/* previous version of PC		*/
inst_t 	 program[MAX_PROG_SIZE];	/* where the instructions are stored	*/
					/* and indexed by the program counter	*/

/* ---------------------------------------------------------------------------- */
oper_t 	  program_counter;		/* the index into the program table.	*/
					/* this is the value the symbol $pc	*/
					/* refers to.				*/
symbol_t program_counter_sym;		/* 					*/
/* ---------------------------------------------------------------------------- */


/* ---------------------------------------------------------------------------- */
int 	  label_cnt;		/* the index into the labels table.		*/
symbol_t* labels[MAX_LABELS];

int 	  id_cnt;		/* */
symbol_t* ids[MAX_IDS];		/* the address information refers the offset	*/
				/* into the memory data strucuture		*/
/* ---------------------------------------------------------------------------- */


/* ---------------------------------------------------------------------------- */
int    mem_addr;		/* the address index into the memory table.	*/
oper_t memory[MAX_MEMORY_SIZE]; /* the memory itself				*/
/* ---------------------------------------------------------------------------- */


/* ---------------------------------------------------------------------------- */
/* This is the older version of the CALL/RETURN instructions implementations.	*/
/* ---------------------------------------------------------------------------- */
oper_t stack1[MAX_STACK_SIZE];	/* stack for PC addresses and return variable	*/
oper_t stack2[MAX_STACK_SIZE];	/* stack for arguments 				*/

int sp1 = 0;			/* stack pointer				*/
int sp2 = 0;			/* stack pointer				*/
/* ---------------------------------------------------------------------------- */


/* ---------------------------------------------------------------------------- */
/* This alternate frame stack is for a more realistic implementation if the 	*/
/* stack frames									*/
/*  Date: Jan. 2012								*/
/* ---------------------------------------------------------------------------- */
// oper_t	frame_stack[MAX_FRAME_STACK_SIZE];
oper_t	frame_pointer;		/* the frame pointer as the offset from the 	*/
				/* beginning of the frame stack data structure	*/
symbol_t frame_pointer_sym;	/* 						*/

oper_t	stack_pointer;		/* used to access fields in the current frame 	*/
symbol_t stack_pointer_sym;	/* 						*/

oper_t	return_address;		/* used to access fields in the current frame 	*/
symbol_t return_address_sym;	/* 						*/

oper_t	 reg[NUMBER_REGISTERS];
symbol_t reg_sym[NUMBER_REGISTERS];

/* ---------------------------------------------------------------------------- */
/* Statistical Variables							*/
/* ---------------------------------------------------------------------------- */

int number_executed_instructions;
int number_memory_instructions;
int number_register_instructions;
int number_arithmetic_instructions;
int number_controlflow_instructions;

/* ---------------------------------------------------------------------------- */
static char *MOD_DATE = 
#include "mod_date.h"
;

int lineno;			/* 						*/
int debug;			/* global debug variable			*/
int lex_debug; 
int yacc_debug; 
int main_debug; 

int step;			/* global step instruction vaiable		*/

/* --------------------------------------------------------------------------- */
#define MAX_FILE_NAME_LENGTH 80

FILE *ifp;
FILE *ofp;
FILE *tfp;
char input_filename[MAX_FILE_NAME_LENGTH];
char output_filename[MAX_FILE_NAME_LENGTH];
char temp_filename[MAX_FILE_NAME_LENGTH];

/* -------------------------------------------------------------------------- */
void helpUsage(int argc, char ** argv){
  printf(" Usage: %s [-debug] [-step] -infile <input filename> -outfile <output filename> [-version <n>]\n",
    argv[0]);
}

/* ---------------------------------------------------------------------------- */
void initStatistics(){
  number_executed_instructions 	  = 0;
  number_memory_instructions 	  = 0;
  number_register_instructions 	  = 0;
  number_arithmetic_instructions  = 0;
  number_controlflow_instructions = 0;
}

/* ---------------------------------------------------------------------------- */
void printStatistics(){
  printf(" ------------------------------------\n");
  printf(" Number Executed Instructions:     %d\n", number_executed_instructions);
  printf(" Number Memory Operations:         %d\n", number_memory_instructions);
  printf(" Number Register Transfers:        %d\n", number_register_instructions);
  printf(" Number Arithmetic Operations:     %d\n", number_arithmetic_instructions);
  printf(" Number Control/Flow Instructions: %d\n", number_controlflow_instructions);
  printf(" ------------------------------------\n");
}

/* -------------------------------------------------------------------------- */
void incrementNumberExecutedInstructions(){
  number_executed_instructions++;
}
void incrementNumberMemoryInstructions(){
  number_memory_instructions++;
}
void incrementNumberRegisterInstructions(){
  number_register_instructions++;
}
void incrementNumberArithmeticInstructions(){
  number_arithmetic_instructions++;
}
void incrementNumberControlFlowInstructions(){
  number_controlflow_instructions++;
}

/* -------------------------------------------------------------------------- */
oper_t *retrieveFirstArgumentFromFrame(){
  int add;
  oper_t *op;

  if(stack_pointer.type != O_INTEGER){
    printf(" *** Error: Incorrect stack pointer (\'$sp\') type. (retrieveFirstArgumentFromFrame)\n");
    return NULL;
  }
  add = stack_pointer.u.ival;
  add += 4;
  op = &memory[add];
  return op;
}

/* -------------------------------------------------------------------------- */
oper_t *retrieveReturnAddressFromFrame(){
  int add;
  oper_t *op;

  if(stack_pointer.type != O_INTEGER){
    printf(" *** Error: Incorrect stack pointer (\'$sp\') type. (retrieveReturnAddressFromFrame)\n");
    return NULL;
  }
  add = stack_pointer.u.ival;
  add += 2;
  op = &memory[add];
  return op;
}

/* -------------------------------------------------------------------------- */
void storeValueAsReturnValueInFrame(oper_t *inop){
  int add;
  oper_t *op;

  if(stack_pointer.type != O_INTEGER){
    printf(" *** Error: Incorrect stack pointer (\'$sp\') type. (retrieveReturnAddressFromFrame)\n");
    return;
  }
  add = stack_pointer.u.ival;
  add += 1;
  op = &memory[add];
  copyOperand(op,inop);
}

/* -------------------------------------------------------------------------- */
void processArgs(int argc, char ** argv){
  int i, n;

  i = 1;
  while(i < argc){

    if(strcmp(argv[i],"-infile") == 0){
      strncpy(input_filename,argv[i+1],MAX_FILE_NAME_LENGTH-1);
      input_filename[MAX_FILE_NAME_LENGTH-1] = '\0';
      ifp = fopen(input_filename,"r");
      if(ifp == NULL){
        printf(" *** Error: Could not open file (%s) as input file (processArgs:main.c)\n",input_filename);
      }
      i += 2;
      continue;
    }

    if(strcmp(argv[i],"-outfile") == 0){
      strncpy(output_filename,argv[i+1],MAX_FILE_NAME_LENGTH-1);
      output_filename[MAX_FILE_NAME_LENGTH-1] = '\0';
      ofp = fopen(output_filename,"w");
      if(ofp == NULL){
        printf(" *** Error: Could not open file (%s) as output file (processArgs:main.c)\n",output_filename);
      }
      i += 2;
      continue;
    }

    if(strcmp(argv[i],"-help") == 0){
      helpUsage(argc,argv);
      i += 1;
      continue;
    }

    if(strcmp(argv[i],"-debug") == 0){
      debug = 1;
      lex_debug = 1;
      yacc_debug = 1;
      main_debug = 1;
      i += 1;
      continue;
    }

    if(strcmp(argv[i],"-debug") == 0){
      step = 1;
      i += 1;
      continue;
    }

    if(strcmp(argv[i],"-version") == 0){
      version = atoi(argv[i+1]);
      if((version != 5) && (version != 6)){
        printf(" *** Error: Invalid version requested. Either 5 or 6 allowed. Default 5 assumed. (processArgs:main.c)\n");
	version = 5;
      }
      i += 2;
      continue;
    }

    i++;
  }

} /* end of processArgs */

/* ----------------------------------------------------------------------------------------- */
int isBlankChar(char c){
  if((c == ' ') || (c == '\t') || (c == '\v') || (c == 0x0a) || (c == 0x0d)){
    return 1;
  }
  return 0;
}

/* ----------------------------------------------------------------------------------------- */
int isEmptyLine(char *line){
  int i, n;
  char c;

  if(line == NULL){
    return 1;
  }
  n = strlen(line);
  if(n <= 0){
    return 1;
  }
  for(i = 0; i < n; i++){
    c = line[i];
    if(c == '\n')
      break;
    if(c == '#')
      break;
    if(isBlankChar(c) == 0)
     return 0;
  }
  return 1;
}

/* ----------------------------------------------------------------------------------------- */
void preprocess(char *ifname, char *tfname){
  int i, len;
  char *str, line1[256], line2[256];
  char *s1, *s2, *s3;
  char *c;

  ifp = fopen(ifname, "r");
  if(ifp == NULL){
    fprintf(stderr, " *** Error: cannot open file (%s) as input ***\n",ifname);
    exit(1);
  }

  tfp = fopen(tfname, "w");
  if(tfp == NULL){
    fprintf(stderr, " *** Error: cannot open file (%s) as temporary output ***\n",tfname);
    exit(1);
  }

  while(fgets(line1, 256, ifp)){
    i = 0;
    c = &line1[0];
    len = strlen(line1);
    if(isEmptyLine(line1)){
      continue;
    }
    while(((*c) != '\0') && (i < 256) && (i < len)){
      if(((*c) == '#') || ((*c) == '@')){
	break;
      }
      if((*c) == '\n'){
	break;
      }
      fputc(*c,tfp);
      c++;
      i++;
    }
    fputc('\n',tfp);
  } /* end of while */
  fclose(ifp);
  fclose(tfp);
}

/* --------------------------------------------------------------------------- */
void checkGotoInstructions(){
  int i, key, found;
  inst_t *ins;
  symbol_t *sym;

  found = 0;
  for(i = 0; i < inst_cnt; i++){
     ins = &(program[i]);
     switch(ins->type){

        case T_GOTO:
	  if(ins->op1.type == O_NULL){
            printf(" *** Error: Empty GOTO instruction target on line %d.\n", ins->line);
	    break;
	  }
          if(ins->op1.type == O_STRING){
  	    sym = lookupItem(label_hash, &found, ins->op1.u.sval);
            if(sym == NULL){
              printf(" *** Error: Label associated with goto in line %d does not exist.\n",
		ins->line);
  	      sym = lookupItem(id_hash, &found, ins->op1.u.sval);
              if(sym != NULL){
                printf(" *** Error: Found target to be a variable.\n");
   	      }
	    }
	  } else {
            printf(" *** Error: Target of GOTO instruction on line %d is not a label.\n",
		ins->line);
	  }
	  break;

        case T_BOOL_COND:
          if(ins->op2.type == O_STRING){
  	    sym = lookupItem(label_hash, &found, ins->op2.u.sval);
            if(sym == NULL){
              printf(" *** Error: Label associated with goto in line %d does not exist.\n",
		ins->line);
  	      sym = lookupItem(id_hash, &found, ins->op2.u.sval);
              if(sym != NULL){
                printf(" *** Error: Found target to be a variable.\n");
   	      }
  	    }
	  } else {
            printf(" *** Error: Target of GOTO instruction on line %d is not a label.\n",
		ins->line);
	  }
	  break;

        case T_COND:
          if(ins->op3.type == O_STRING){
  	    sym = lookupItem(label_hash, &found, ins->op3.u.sval);
            if(sym == NULL){
              printf(" *** Error: Label associated with goto in line %d does not exist.\n",
		ins->line);
  	      sym = lookupItem(id_hash, &found, ins->op3.u.sval);
              if(sym != NULL){
                printf(" *** Error: Found target to be a variable.\n");
   	      }
  	    }
	  } else {
            printf(" *** Error: Target of GOTO instruction on line %d is not a label.\n",
		ins->line);
	  }
	  break;

        default:
	  break;
     }

  } /* end of for on instructions */
}

/* --------------------------------------------------------------------------- */
void mapVariablesInInstruction(inst_t *ins){
  int addr;
  symbol_t* sym;

  if(ins == NULL){
    return;
  }

  switch(ins->type){
    case T_OPER:
    case T_UNARY:
      if(ins->op1.type == O_STRING){
        sym = getSymbolRef(ins->op1.u.sval);
      }
      if(ins->op2.type == O_STRING){
        sym = getSymbolRef(ins->op2.u.sval);
      }
      if(ins->op3.type == O_STRING){
        sym = getSymbolRef(ins->op3.u.sval);
      }
      break;
    case T_COPY:
      if(ins->op1.type == O_STRING){
        sym = getSymbolRef(ins->op1.u.sval);
      }
      if(ins->op2.type == O_STRING){
        sym = getSymbolRef(ins->op2.u.sval);
      }
      break;
    case T_GOTO:
      break;
    case T_PUTPAR:
      if(version == 6){
        printf(" *** Error: GETPAR/PUTPAR instruction not supported in version 6. Instruction ignored\n");
        break;
      }
      if(ins->op1.type == O_STRING){
        sym = getSymbolRef(ins->op1.u.sval);
      }
      break;
    case T_GETPAR:
      if(version == 6){
        printf(" *** Error: GETPAR/PUTPAR instruction not supported in version 6. Instruction ignored\n");
        break;
      }
      if(ins->op1.type == O_STRING){
        sym = getSymbolRef(ins->op1.u.sval);
      } else {
        printf(" *** Error: Argument of getparam instruction in line %d must be a variable\n",ins->line);
      }
      break;
    case T_ADDRESSOF:
      if(ins->op1.type == O_STRING){
        sym = getSymbolRef(ins->op1.u.sval);
        if(isRegisterSymbol(sym)){
          printf(" *** Error: Target argument of addressof instruction in line %d cannot be a register\n",ins->line);
	}
      } else {
        printf(" *** Error: Target argument of addressof instruction in line %d must be a variable\n",ins->line);
      }
      if(ins->op2.type == O_STRING){
        sym = getSymbolRef(ins->op2.u.sval);
        if(isRegisterSymbol(sym)){
          printf(" *** Error: Source argument of addressof instruction in line %d cannot be a register(1)\n",ins->line);
	}
      } else {
        printf(" *** Error: Source argument of addressof instruction in line %d must be a variable\n",ins->line);
      }
      break;
    case T_DEREF:
      if(ins->op1.type == O_STRING){
        sym = getSymbolRef(ins->op1.u.sval);
      } else {
        printf(" *** Error: Argument of deref instruction in line %d must be a variable or register\n",ins->line);
      }
      if(ins->op2.type == O_STRING){
        sym = getSymbolRef(ins->op2.u.sval);
      } else {
        printf(" *** Error: Argument of deref instruction in line %d must be a variable\n",ins->line);
      }
      break;
    case T_PTRASSIGN:
      if(ins->op1.type == O_STRING){
        sym = getSymbolRef(ins->op1.u.sval);
      } else {
        printf(" *** Error: Argument of pointer assignment instruction in line %d must be a variable\n",ins->line);
      }
      break;
    default:
      break;
  }
}

/* --------------------------------------------------------------------------- */
void flattenStringInMemory(int addr, char *str){
  int i, len;
  oper_t *dst_op, *src_op;

  len = strlen(str);
  for(i = 0; i < len; i++){
    dst_op = &(memory[addr+i]);
    src_op = createCharValOperand(str[i]);
    copyOperand(dst_op,src_op);
  }
  dst_op = &(memory[addr+len]);
  src_op = createCharValOperand('\0');
  copyOperand(dst_op,src_op);
}

/* --------------------------------------------------------------------------- */
void initVariable(symbol_t *sym){
  int addr;
  oper_t *dst_op;

  if(sym == NULL)
    return;

  dst_op = &(memory[sym->address]);
  if(sym->initvalue != NULL){
    if(sym->initvalue->type == O_STRING){
     flattenStringInMemory(sym->address,sym->initvalue->u.sval);
    } else {
      copyOperand(dst_op,sym->initvalue);
    }
  } else {
    dst_op->type = O_NULL;
  }
}

void initVariables(){
  int i;
  symbol_t *sym;
  for(i = 0; i < id_cnt; i++) {
    sym = ids[i];
    if(sym != NULL){
      initVariable(sym);
    }
  }
}

/* --------------------------------------------------------------------------- */
void initRegisters(){
  int i; 

  setOperandInteger(&program_counter, 0);
  initSymbol(&program_counter_sym, REG_SYM,"$pc",0,0,0);
  setOperandInteger(&frame_pointer, MAX_MEMORY_SIZE-1);
  initSymbol(&frame_pointer_sym, REG_SYM,"$fp",0,0,0);
  setOperandInteger(&return_address, 0);
  initSymbol(&return_address_sym, REG_SYM,"$ra",0,0,0);
  setOperandInteger(&stack_pointer, 0);
  initSymbol(&stack_pointer_sym, REG_SYM,"$sp",0,0,0);

  for(i = 0; i < NUMBER_REGISTERS; i++){
    setOperandInteger(&reg[i], 0);
    initSymbol(&reg_sym[i], REG_SYM,"$rXX",0,0,i);
  }
}

/* --------------------------------------------------------------------------- */
void mapVariables(){
  int i;
  inst_t *ins;

  for(i = 0; i < inst_cnt; i++) {
    ins = &(program[i]);
    mapVariablesInInstruction(ins);
  }
}

/* --------------------------------------------------------------------------- */
void dumpLabelTable(){
  int i;
  symbol_t *lab;

  fprintf(stdout,"\n*** Label Table Dump ***\n");
  for(i=0; i < label_cnt; i++) {
    lab = labels[i];
    printf("%s = %d\n",lab->tag,lab->address);
  }
}

/* --------------------------------------------------------------------------- */
void dumpSymbolTable(){
  int i;
  symbol_t *ins;

  fprintf(stdout,"\n*** Symbol Table Dump ***\n");
  for(i=0; i < id_cnt; i++) {
    ins = ids[i];
    printf("[%x] ",(int)ins);
    printSymbol(ins);
    printf(" : ");
    printOperand(&memory[ins->address]);
    printf("\n");
  }
}

/* --------------------------------------------------------------------------- */
void dumpMemory(){
  int i;
  oper_t *op;

  fprintf(stdout,"\n*** Memory Dump ***\n");
  for(i=0; i < mem_addr; i++) {
    op = &(memory[i]);
    printf("[%d] : ",i);
    printOperand(op);
    printf("\n");
  }
}

/* --------------------------------------------------------------------------- */
void dumpFrameStack(){
  int i, n;
  oper_t *op;

  fprintf(stdout,"\n*** Frame Stack Dump ***\n");
  n = frame_pointer.u.ival;
  fprintf(stdout," Frame Pointer: %d\n",n);
  n = stack_pointer.u.ival;
  for(i = MAX_MEMORY_SIZE-1; i >= n; i--){
    op = &(memory[i]);
    printf("[%d] : ",i);
    printOperand(op);
    printf("\n");
  }
}

/* --------------------------------------------------------------------------- */
void dumpSpecialRegisters(){
  printf("[$pc]  : ");
  printOperand(&program_counter);
  printf("\n");
  printf("[$sp]  : ");
  printOperand(&stack_pointer);
  printf("\n");
  printf("[$fp]  : ");
  printOperand(&frame_pointer);
  printf("\n");
  printf("[$ra]  : ");
  printOperand(&return_address);
  printf("\n");
}

/* --------------------------------------------------------------------------- */
void dumpRegisters(){
  int i;
  oper_t *op;

  fprintf(stdout,"\n*** Registers Dump ***\n");
  dumpSpecialRegisters();

  for(i = 0; i < NUMBER_REGISTERS; i++){
    op = &(reg[i]);
    printf("[$r%2d] : ",i);
    printOperand(op);
    printf("\n");
  }

}

/* --------------------------------------------------------------------------- */
oper_t* getRegisterVariableReference(int id){
printf(" -*******- NOT USED -******-\n");
  if(id == PC_REGISTER){
    return &(program_counter);
  }
  if(id == SP_REGISTER){
    return &(stack_pointer);
  }
  if(id == FP_REGISTER){
    return &(frame_pointer);
  }
  if(id == RA_REGISTER){
    return &(return_address);
  }
  if((id < 0) || (id > 31)){ 
    return NULL;
  }
  return &(reg[id]);
}

/* --------------------------------------------------------------------------- */
int main(int argc, char *argv[]) {
  int val, idx;
  symbol_t *sym;

  step  	= 0;

  lex_debug  	= 0;
  yacc_debug 	= 0;
  main_debug 	= 0;
  lineno  	= 0;

  mem_addr 	= 0;

  inst_cnt 	= 0;
  label_cnt 	= 0;
  id_cnt 	= 0;

  version 	= 5;

  initStatistics();

  fprintf(stdout,"C3E Simulator %s (ver. 6)\n",MOD_DATE);
  if(argc < 4){
    helpUsage(argc, argv);
    exit(1);
  }

  processArgs(argc, argv);

  snprintf(temp_filename,80,"temp_%s",input_filename);
  preprocess(input_filename, temp_filename);

  label_hash = createHashTable(256);
  id_hash    = createHashTable(1024);
  fct_hash   = createHashTable(256);

  sym = getLabelRef("exit");
  sym->address = EXIT_ADDRESS;

  sym = getLabelRef("cread");
  sym->address = CREAD_ADDRESS;

  sym = getLabelRef("cwrite");
  sym->address = CWRITE_ADDRESS;

  sym = getLabelRef("iread");
  sym->address = IREAD_ADDRESS;

  sym = getLabelRef("iwrite");
  sym->address = IWRITE_ADDRESS;

  sym = getLabelRef("fread");
  sym->address = FREAD_ADDRESS;

  sym = getLabelRef("fwrite");
  sym->address = FWRITE_ADDRESS;

  yyin = fopen(temp_filename,"r");
  if(yyin == NULL) {
    fprintf(stderr, " *** Error opening temp file (%s) as input. Exiting.\n", temp_filename);
    return 1;
  }

  if(main_debug)
    fprintf(stdout," > Begin parsing \n");
  yyparse();
  if(main_debug)
    fprintf(stdout," > End parsing \n");

  if(main_debug)
    fprintf(stdout," > Begin checks \n");

  initRegisters(); 
  initVariables(); 
  checkGotoInstructions();
  if(main_debug)
    fprintf(stdout," > End checks \n");


  if(main_debug)
    fprintf(stdout," > Begin simulation \n");
  if(version == 5){
    fprintf(stdout," > Selected version V5 \n");
    simulateV5();
  } else {
    fprintf(stdout," > Selected version V6 \n");
    simulateV6();
  }
  if(main_debug)
    fprintf(stdout," > End simulation \n");

  printStatistics();
  return 0;
}

/* ---------------------------------------------------------------------------- */
/* Function: simulate								*/
/* ---------------------------------------------------------------------------- */
/* Description:									*/
/* ---------------------------------------------------------------------------- */
void simulateV5() {
  int i, n;
  symbol_t *ins, *lab;
  char c;

  lab = getLabelRef("main");
  if(lab == NULL){
    printf(" *** Error: Could not find the starting label \"main\" in the code\n"); 
    return;
  }
  curr_pc = lab->address;
  printf(" > MAIN: Found main at pc = %d\n",curr_pc); 
  while(1) {
    if(main_debug){
      printf(" > MAIN: Executing instructon at pc = %d\n",curr_pc); 
    }
    curr_pc = exec_instruction (curr_pc);

    incrementNumberExecutedInstructions();

    if(step){
      printf(" > Press \'return\' key to continue... "); 
      scanf(&c);
    }

    if (curr_pc == -1){
      break;
    }
  }

  if(main_debug){
    printf(" > MAIN: Terminated execution with instructon at pc = %d\n",curr_pc); 
  }
  dumpSymbolTable();
  dumpLabelTable();
  dumpMemory();
  dumpFrameStack();
  dumpRegisters();

} /* --- end of simulate ver. 5 --- */

void simulateV6() {
  int i, n;
  symbol_t *ins, *lab;
  char c;

  lab = getLabelRef("main");
  if(lab == NULL){
    printf(" *** Error: Could not find the starting label \"main\" in the code\n"); 
    return;
  }
  curr_pc = 0;
  setOperandInteger(&program_counter, lab->address);
  printf(" > MAIN: Found main at $pc = %d\n",getIntegerOperandValue(&program_counter)); 
  while(1) {
    if(main_debug){
      printf(" > MAIN: Executing instructon at $pc = %d\n", getIntegerOperandValue(&program_counter)); 
    }
    n = exec_instruction(getIntegerOperandValue(&program_counter)); 
    setOperandInteger(&program_counter, n);

    if(step){
      dumpSpecialRegisters();

      printf(" > Press \'return\' key to continue... "); 
      scanf(&c);
    }

    incrementNumberExecutedInstructions();

    if(n < 0){
      if(n == EXIT_ADDRESS){
        printf(" > MAIN: Normal execution completion via Monitor.\n");
        break;
      }
      if(exec_io_proc(n) == 1){
        printf(" > MAIN: Exiting execution via Trap in IO Function.\n");
        break;
      }
    }
  } /* --- main simulation loop --- */

  if(main_debug){
    printf(" > MAIN: Terminated execution with instructon at $pc = %d\n",
	getIntegerOperandValue(&program_counter)); 
  }

  dumpSymbolTable();
  dumpLabelTable();
  dumpMemory();
  dumpFrameStack();
  dumpRegisters();
  dumpSpecialRegisters();

  if(main_debug){
  }
} /* --- end of simulate version 6 --- */

/* ---------------------------------------------------------------------------- */
/* Need to fix this for the case where the symbol is a register			*/
/* ---------------------------------------------------------------------------- */
oper_t *getSourceOperandValue(oper_t *src){
  int addr;
  symbol_t* sym;
  oper_t *res;

  if(src == NULL){
    printf(" *** Error: Null source operand. Null value returned.\n");
    return NULL;
  }
  if(src->type == O_STRING){
    if(isRegisterName(src->u.sval)){
     res = getRegisterSymbolOperand(src->u.sval);
    } else {
      sym = getSymbolRef(src->u.sval);
      if(sym == NULL){
        printf(" *** Warn: Operand (%s) is an unmapped variable. Null value returned.\n",src->u.sval);
        res = NULL;
      } else {
        addr = sym->address;
        res = &(memory[addr]);
      }
    }
  } else {
    res = src;
  }
  return res;
}

/* --------------------------------------------------------------------------- */
void implementBinaryPlusOperator(oper_t *d, oper_t *s1, oper_t *s2){
  int ival;
  double dval;
  char cval;

  if((s1->type == O_INTEGER) && (s2->type == O_INTEGER)){
    ival = s1->u.ival + s2->u.ival;
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_INTEGER) && (s2->type == O_DOUBLE)){
    dval = (1.0*s1->u.ival) + s2->u.dval;
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    dval = (s1->u.dval) + (1.0*s2->u.ival);
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    dval = (s1->u.dval) + (s2->u.dval);
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  printf(" *** Warn: Incompatible types in PLUS operator. Instruction ignored\n");
  return;
}

/* --------------------------------------------------------------------------- */
void implementUnaryMinusOperator(oper_t *d, oper_t *s1){
  int ival;
  double dval;

  if(s1->type == O_INTEGER){
    ival = -(s1->u.ival);
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if(s1->type == O_DOUBLE){
    dval = -(s1->u.dval);
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  printf(" *** Warn: Incompatible type in UMINUS operator. Instruction ignored\n");
  return;
}

/* --------------------------------------------------------------------------- */
void implementBinaryMinusOperator(oper_t *d, oper_t *s1, oper_t *s2){
  int ival;
  double dval;
  char cval;

  if((s1->type == O_INTEGER) && (s2->type == O_INTEGER)){
    ival = s1->u.ival - s2->u.ival;
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_INTEGER) && (s2->type == O_DOUBLE)){
    dval = (1.0*s1->u.ival) - s2->u.dval;
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    dval = (s1->u.dval) - (1.0*s2->u.ival);
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    dval = (s1->u.dval) - (s2->u.dval);
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  printf(" *** Warn: Incompatible types in MINUS operator. Instruction ignored\n");
  return;
}

/* --------------------------------------------------------------------------- */
void implementBinaryTimesOperator(oper_t *d, oper_t *s1, oper_t *s2){
  int ival;
  double dval;
  char cval;

  if((s1->type == O_INTEGER) && (s2->type == O_INTEGER)){
    ival = s1->u.ival * s2->u.ival;
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_INTEGER) && (s2->type == O_DOUBLE)){
    dval = (1.0*s1->u.ival) * s2->u.dval;
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    dval = (s1->u.dval) * (1.0*s2->u.ival);
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    dval = (s1->u.dval) * (s2->u.dval);
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  printf(" *** Warn: Incompatible types in TIMES operator. Instruction ignored\n");
  return;
}

/* --------------------------------------------------------------------------- */
void implementBinaryDivisionOperator(oper_t *d, oper_t *s1, oper_t *s2){
  int ival;
  double dval;
  char cval;

  if((s1->type == O_INTEGER) && (s2->type == O_INTEGER)){
    if((s1->u.ival % s2->u.ival) == 0){
      ival = s1->u.ival / s2->u.ival;
      d->u.ival = ival;
      d->type = O_INTEGER;
    } else {
      dval = (1.0*s1->u.ival) / (1.0*s2->u.ival);
      d->u.dval = dval;
      d->type = O_DOUBLE;
    }
    return;
  }
  if((s1->type == O_INTEGER) && (s2->type == O_DOUBLE)){
    dval = (1.0*s1->u.ival) / s2->u.dval;
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    dval = (s1->u.dval) / (1.0*s2->u.ival);
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    dval = (s1->u.dval) / (s2->u.dval);
    d->u.dval = dval;
    d->type = O_DOUBLE;
    return;
  }
  printf(" *** Warn: Incompatible types in DIV operator. Instruction ignored\n");
  return;
}

/* --------------------------------------------------------------------------- */
void implementBinaryBitwiseAndOperator(oper_t *d, oper_t *s1, oper_t *s2){
  int ival;
  double dval;
  char cval;

  if((s1->type == O_INTEGER) && (s2->type == O_INTEGER)){
    ival = s1->u.ival & s2->u.ival;
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_INTEGER) && (s2->type == O_DOUBLE)){
    printf(" *** Warn: Converting second double operand of bitwise AND operator.\n");
    ival = (s1->u.ival) & (int)(s2->u.dval);
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    printf(" *** Warn: Converting first double operand of bitwise AND operator.\n");
    ival = (int)(s1->u.dval) & (s2->u.ival);
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_DOUBLE)){
    printf(" *** Warn: Converting both double operands of bitwise AND operator.\n");
    ival = (int)(s1->u.dval) & (int)(s2->u.dval);
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  printf(" *** Warn: Incompatible types in bitwise AND operator. Instruction ignored\n");
  return;
}

/* --------------------------------------------------------------------------- */
void implementUnaryLogicalNotOperator(oper_t *d, oper_t *s1){
  if((d == NULL) || (s1 == NULL)){
    return;
  }
  switch(s1->type){
    case O_INTEGER:
      if(s1->u.ival != 0){
        d->u.ival = 0;
      } else {
        d->u.ival = 1;
      }
      d->type = O_INTEGER;
      break;
    case O_DOUBLE:
      if(s1->u.dval != 0.0){
        d->u.ival = 0;
      } else {
        d->u.ival = 1;
      }
      d->type = O_INTEGER;
      break;
    case O_CHAR:
      if(s1->u.cval != '\0'){
        d->u.ival = 0;
      } else {
        d->u.ival = 1;
      }
      d->type = O_INTEGER;
      break;
    case O_STRING:
      printf(" *** Warn: Incompatible types in logical NOT operator. Instruction ignored\n");
      break;
    default:
      break;
  }
}

/* --------------------------------------------------------------------------- */
void implementUnaryBitwiseNotOperator(oper_t *d, oper_t *s1){
  if((d == NULL) || (s1 == NULL)){
    return;
  }
  switch(s1->type){
    case O_INTEGER:
      d->u.ival = ~(s1->u.ival);
      d->type = O_INTEGER;
      break;
    case O_DOUBLE:
      printf(" *** Warn: Incompatible types in bitwise NOT operator. Instruction ignored\n");
      break;
    case O_CHAR:
      d->u.cval = (char)(~s1->u.cval);
      d->type = O_CHAR;
      break;
    case O_STRING:
      printf(" *** Warn: Incompatible types in bitwise NOT operator. Instruction ignored\n");
      break;
    default:
      break;
  }
}

/* --------------------------------------------------------------------------- */
void implementBinaryLogicalAndOperator(oper_t *d, oper_t *s1, oper_t *s2){
  int ival;
  double dval;
  char cval;

  if((s1->type == O_INTEGER) && (s2->type == O_INTEGER)){
    if((s1->u.ival != 0) && (s2->u.ival != 0)){
      d->u.ival = 1;
    } else {
      d->u.ival = 0;
    }
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_INTEGER) && (s2->type == O_DOUBLE)){
    printf(" *** Warn: Converting second double operand of logical AND operator.\n");
    if((s1->u.ival != 0) && (s2->u.dval != 0.0)){
      d->u.ival = 1;
    } else {
      d->u.ival = 0;
    }
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    printf(" *** Warn: Converting first double operand of logical AND operator.\n");
    if((s1->u.dval != 0.0) && (s2->u.ival != 0)){
      d->u.ival = 1;
    } else {
      d->u.ival = 0;
    }
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_DOUBLE)){
    printf(" *** Warn: Converting both double operands of logical AND operator.\n");
    if((s1->u.dval != 0.0) && (s2->u.dval != 0.0)){
      d->u.ival = 1;
    } else {
      d->u.ival = 0;
    }
    d->type = O_INTEGER;
    return;
  }
  printf(" *** Warn: Incompatible types in logical AND operator. Instruction ignored\n");
  return;
}

/* --------------------------------------------------------------------------- */
void implementBinaryBitwiseOrOperator(oper_t *d, oper_t *s1, oper_t *s2){
  int ival;
  double dval;
  char cval;

  if((s1->type == O_INTEGER) && (s2->type == O_INTEGER)){
    ival = s1->u.ival | s2->u.ival;
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_INTEGER) && (s2->type == O_DOUBLE)){
    printf(" *** Warn: Converting second double operand of bitwise OR operator.\n");
    ival = (s1->u.ival) | (int)(s2->u.dval);
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    printf(" *** Warn: Converting first double operand of bitwise OR operator.\n");
    ival = (int)(s1->u.dval) | (s2->u.ival);
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_DOUBLE)){
    printf(" *** Warn: Converting both double operands of bitwise OR operator.\n");
    ival = (int)(s1->u.dval) | (int)(s2->u.dval);
    d->u.ival = ival;
    d->type = O_INTEGER;
    return;
  }
  printf(" *** Warn: Incompatible types in bitwise OR operator. Instruction ignored\n");
  return;
}

/* --------------------------------------------------------------------------- */
void implementBinaryLogicalOrOperator(oper_t *d, oper_t *s1, oper_t *s2){
  int ival;
  double dval;
  char cval;

  if((s1->type == O_INTEGER) && (s2->type == O_INTEGER)){
    if((s1->u.ival != 0) || (s2->u.ival != 0)){
      d->u.ival = 1;
    } else {
      d->u.ival = 0;
    }
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_INTEGER) && (s2->type == O_DOUBLE)){
    printf(" *** Warn: Converting second double operand of logical OR operator.\n");
    if((s1->u.ival != 0) || (s2->u.dval != 0.0)){
      d->u.ival = 1;
    } else {
      d->u.ival = 0;
    }
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_INTEGER)){
    printf(" *** Warn: Converting first double operand of logical OR operator.\n");
    if((s1->u.dval != 0.0) || (s2->u.ival != 0)){
      d->u.ival = 1;
    } else {
      d->u.ival = 0;
    }
    d->type = O_INTEGER;
    return;
  }
  if((s1->type == O_DOUBLE) && (s2->type == O_DOUBLE)){
    printf(" *** Warn: Converting both double operands of logical OR operator.\n");
    if((s1->u.dval != 0.0) || (s2->u.dval != 0.0)){
      d->u.ival = 1;
    } else {
      d->u.ival = 0;
    }
    d->type = O_INTEGER;
    return;
  }
  printf(" *** Warn: Incompatible types in logical OR operator. Instruction ignored\n");
  return;
}

/* ---------------------------------------------------------------------------- */
symbol_t* checkGotoTargetInstruction(inst_t *ins, int op_id){
  int addr, found;
  oper_t *op;
  symbol_t *sym;

  if(ins == NULL){
    return NULL;
  }

  if(op_id == 1){
    op = &(ins->op1);
  } else {
    if(op_id == 2){
      op = &(ins->op2);
    } else {
      op = &(ins->op3);
    }
  }

  found = 0;
  if(op->type == O_STRING){
    sym = lookupItem(label_hash, &found, op->u.sval);
    if(sym == NULL){
      printf(" *** Error: Label associated with goto instruction in line %d does not exist. Instruction ignored.\n",
 	ins->line);
      return NULL;
    } else {
      return sym;
    }
  } else {
    printf(" *** Error: No Symbol associated with goto instruction in line %d; ignored.\n",ins->line);
    return NULL;
  }
}

/* ---------------------------------------------------------------------------- */
int implementGotoInstruction(int pc, inst_t *ins){
  symbol_t *sym;

  sym = checkGotoTargetInstruction(ins, 1);
  if(sym != NULL){
    pc = sym->address;
  } else {
    pc = pc + 1;
  }
  return pc;
}

/* ---------------------------------------------------------------------------- */
int implementBooleanCondInstruction(int pc, inst_t *ins){
  int addr, found, goto_pc, tmp;
  symbol_t *sym;
  oper_t *dst_op;

  if(ins == NULL)
    return (pc+1);

  sym = checkGotoTargetInstruction(ins, 2);
  if(sym != NULL){
    goto_pc = sym->address;
  } else {
    return (pc+1);
  }

  /* --- valid target label --- */
  switch(ins->op1.type){
    case O_STRING:
      sym = lookupItem(id_hash, &found, ins->op1.u.sval);
      if(sym == NULL){
        printf(" *** Error: No symbol associated with predicate in instruction in line %d. Instruction ignored.\n", ins->line);
	return (pc+1);
      } else {
	addr = sym->address;	/* --- address of the symbol to fetch operand --- */
	dst_op = &(memory[addr]);
	if(dst_op->type == O_INTEGER){
	  tmp = (dst_op->u.ival != 0 ? 1 : 0);
   	} else {
	  if(dst_op->type == O_DOUBLE){
	    tmp = (dst_op->u.dval != 0.0 ? 1 : 0);
	  } else {
	    if(dst_op->type == O_CHAR){
	      tmp = (dst_op->u.cval != '\0' ? 1 : 0);
	    } else {
              printf(" *** Error: Symbol associated with predicate has non-compliant type value (string). Instruction ignored.\n");
	      tmp = 0;
	    }
	  }
   	}
	if(tmp != 0){
	  pc = goto_pc;
	} else {
	  pc = pc + 1;
	}
      }
      break;
    case O_INTEGER:
      if(ins->op1.u.ival != 0){
	pc = goto_pc;
      } else {
	pc = pc + 1;
      }
      break;
    case O_DOUBLE:
      if(ins->op1.u.dval != 0.0){
	pc = goto_pc;
      } else {
        pc = pc + 1;
      }
      break;
    case O_CHAR:
      if(ins->op1.u.cval != '\0'){
	pc = goto_pc;
      } else {
        pc = pc + 1;
      }
      break;
    default:
      printf(" *** Error: Invalid operand type in conditional instruction in line %d. Instruction ignored.\n", ins->line);
      pc = pc + 1;
      break;
  }
  return pc;
}

/* ---------------------------------------------------------------------------- */
oper_t *dereferenceStringOper(inst_t *ins, oper_t *op){
  int addr, found;
  symbol_t *sym;
  oper_t *res;

  if((op == NULL) || (op->type != O_STRING)){
    return NULL;
  }

  if(isRegisterName(op->u.sval)){
    res = getRegisterSymbolOperand(op->u.sval);
    return res;
  }

  sym = lookupItem(id_hash, &found, op->u.sval);
  if(sym == NULL){
    printf(" *** Error: Cannot find mapping for symbol (%s) in instruction in line %d. Instruction ignored.\n", op->u.sval,ins->line);
    return NULL;
  } else {        
    addr = sym->address;    /* --- address of the symbol to fetch operand --- */
    return &(memory[addr]);
  }
}

/* ---------------------------------------------------------------------------- */
symbol_t *checkStringOperSymbol(inst_t *ins, oper_t *op){
  int addr, found;
  symbol_t *sym;
  oper_t *res;

  if((op == NULL) || (op->type != O_STRING)){
    return NULL;
  }

  if(isRegisterName(op->u.sval)){
    res = getRegisterSymbolOperand(op->u.sval);
    return res;
  }

  sym = lookupItem(id_hash, &found, op->u.sval);
  if(sym == NULL){
    printf(" *** Error: Cannot find mapping for symbol (%s) in instruction in line %d. Instruction ignored.\n", op->u.sval,ins->line);
    return NULL;
  } else {        
    return sym;
  }
}

/* ---------------------------------------------------------------------------- */
oper_t *evalRHSInstructionOperand(inst_t *ins, oper_t *op, char *str1){
  int addr, found;
  oper_t *res;
  symbol_t *sym;

  if(op->type == O_STRING){
    if(isRegisterName(op->u.sval)){
     res = getRegisterSymbolOperand(op->u.sval);
     return res;
    }
    sym = lookupItem(id_hash, &found, op->u.sval);
    if(sym == NULL){
      printf(" *** Error: RHS of %s is an unmapped variable. Instruction ignored.\n",str1);
      return NULL;
    }
    addr = sym->address;
    res = &(memory[addr]);
    if(res->type == O_NULL){
      printf(" *** Warn: RHS of %s is an un-initialized variable. Possible erroneous program behavior.\n",str1);
    }
  } else {
    res = op;
  }
  return res;
}

/* ---------------------------------------------------------------------------- */
int checkUninitializedValue(inst_t *ins, oper_t *op){

  if((ins == NULL) || (op == NULL)){
    return 1;
  }
  if(op->type == O_NULL){
    printf(" *** Error: Operand in instruction in line %d is uninitialized. Instruction ignored.\n",ins->line);
    return 1;
  }
  return 0;
}

/* ---------------------------------------------------------------------------- */
int evaluatePredicate(inst_t *ins, oper_t* op1, int oper, oper_t *op2){
  int tmp, res;
  oper_t *lop1, *lop2;

  if(ins == NULL){
    return 0;
  }
  if((op1 == NULL) || (op2 == NULL)){
    printf(" *** Error: Invalid operands in conditional instruction in line %d. Instruction ignored\n",
     ins->line);
    return 0;
  }

  res = 0;
  if(op1->type == O_STRING){
    lop1  = dereferenceStringOper(ins, op1);
  } else {
    lop1 = op1;
  }
  if(checkUninitializedValue(ins, lop1) == 1)
    return 0;
  if(op2->type == O_STRING){
    lop2  = dereferenceStringOper(ins, op2);
  } else {
    lop2 = op2;
  }
  if(checkUninitializedValue(ins, lop2) == 1)
    return 0;

  /* --- need to dereference the operands if they are variables first , then evaluate predicates --- */

  if(lop1->type != lop2->type){
    printf(" *** Error: Type mismatch of operands in conditional instruction in line %d. Instruction ignored\n",ins->line);
    return 0;
  } else {
    /* --- type is the same so we can compare values --- */
    res = 0; 
    switch(lop1->type){
      case O_INTEGER:
        switch(oper) {
          case LT:  res = (lop1->u.ival < lop2->u.ival  ? 1 : 0); break;
          case GT:  res = (lop1->u.ival > lop2->u.ival  ? 1 : 0); break;
          case LE:  res = (lop1->u.ival <= lop2->u.ival ? 1 : 0); break;
          case GE:  res = (lop1->u.ival >= lop2->u.ival ? 1 : 0); break;
          case EQ:  res = (lop1->u.ival == lop2->u.ival ? 1 : 0); break;
          case NEQ: res = (lop1->u.ival != lop2->u.ival ? 1 : 0); break;
	  default:  
    	    printf(" *** Error: Invalid relational operator in conditional instruction in line %d. Instruction ignored\n",ins->line);
	    break;
        }
        break;
      case O_DOUBLE:
        switch(oper) {
          case LT:  res = (lop1->u.dval < lop2->u.dval  ? 1 : 0); break;
          case GT:  res = (lop1->u.dval > lop2->u.dval  ? 1 : 0); break;
          case LE:  res = (lop1->u.dval <= lop2->u.dval ? 1 : 0); break;
          case GE:  res = (lop1->u.dval >= lop2->u.dval ? 1 : 0); break;
          case EQ:  res = (lop1->u.dval == lop2->u.dval ? 1 : 0); break;
          case NEQ: res = (lop1->u.dval != lop2->u.dval ? 1 : 0); break;
	  default:  
    	    printf(" *** Error: Invalid relational operator in conditional instruction in line %d. Instruction ignored\n",ins->line);
	    break;
        }
        break;
      case O_CHAR:
        switch(oper) {
          case LT:  res = (lop1->u.cval < lop2->u.cval  ? 1 : 0); break;
          case GT:  res = (lop1->u.cval > lop2->u.cval  ? 1 : 0); break;
          case LE:  res = (lop1->u.cval <= lop2->u.cval ? 1 : 0); break;
          case GE:  res = (lop1->u.cval >= lop2->u.cval ? 1 : 0); break;
          case EQ:  res = (lop1->u.cval == lop2->u.cval ? 1 : 0); break;
          case NEQ: res = (lop1->u.cval != lop2->u.cval ? 1 : 0); break;
	  default:  
    	    printf(" *** Error: Invalid relational operator in conditional instruction in line %d. Instruction ignored\n",ins->line);
	    break;
        }
        break;
      default:
    	printf(" *** Error: Invalid types for operands in conditional instruction in line %d. Instruction ignored\n",ins->line);
	res = 0;
        break;
    }
  }
  return res;
}

/* ---------------------------------------------------------------------------- */
int implementPredCondInstruction(int pc, inst_t *ins){
  int addr, found, goto_pc, tmp;
  symbol_t *sym;
  oper_t *dst_op;

  if(ins == NULL)
    return (pc+1);

  sym = checkGotoTargetInstruction(ins, 3);
  if(sym != NULL){
    goto_pc = sym->address;
  } else {
    return (pc+1);
  }

  tmp = evaluatePredicate(ins, &(ins->op1), ins->operator, &(ins->op2));
  if(tmp != 0){
    pc = goto_pc;
  } else {
    pc = pc + 1;
  }
  return pc;
}

/* ---------------------------------------------------------------------------- */
symbol_t* getAddressableSymbolRef(char *st, inst_t* ins, char *str2){
  int found;
  symbol_t *sym;

  sym = lookupItem(id_hash, &found, st);
  if(sym != NULL){
     return sym;
  }
  sym = lookupItem(fct_hash, &found, st);
  if(sym != NULL){
     return sym;
  }
  sym = lookupItem(label_hash, &found, st);
  if(sym != NULL){
     return sym;
  }
  return NULL;
}

oper_t* checkOperandInstructionVariable(inst_t* ins, int op_id, char *str1, char *str2){
  int addr, found;
  symbol_t *sym;
  oper_t *op, *res;

  if(ins == NULL){
    return NULL;
  }

  if(op_id == 1){
    op = &(ins->op1);
  } else {
    if(op_id == 2){
      op = &(ins->op2);
    } else {
      op = &(ins->op3);
    }
  }

  found = 0;
  if(op->type == O_STRING){
    if(isRegisterName(op->u.sval)){
     res = getRegisterSymbolOperand(op->u.sval);
     return res;
    } else {
      sym = lookupItem(id_hash, &found, op->u.sval);
      if(sym == NULL){
        printf(" *** Error: %s of %s instruction on line %d is an unmapped variable. Instruction ignored.\n",
        str1, str2, ins->line);
        return NULL;
      }
      addr = sym->address;
      return (&(memory[addr]));
    }
  }
  printf(" *** Error: %s of %s instruction on line %d is not a variable. Instruction ignored.\n",
    str1, str2, ins->line);
    return NULL;
}

/* ---------------------------------------------------------------------------- */
symbol_t* checkTargetFunctionSymbol(inst_t* ins, int op_id, char *str1, char *str2){
  int addr, found;
  symbol_t *sym;
  oper_t *op;

  if(ins == NULL){
    return NULL;
  }

  if(op_id == 1){
    op = &(ins->op1);
  } else {
    if(op_id == 2){
      op = &(ins->op2);
    } else {
      op = &(ins->op3);
    }
  }

  found = 0;
  if(op->type == O_STRING){
    sym = lookupItem(label_hash, &found, op->u.sval);
    if(sym == NULL){
      printf(" *** Error: %s of %s call instruction on line %d is an invalid function name. Instruction ignored.\n",
        str1, str2, ins->line);
      return NULL;
    }
/*    addr = sym->address; 
    return (labels[addr]);
*/
    return sym;
  } else {
    printf(" *** Error: %s of %s call instruction on line %d is not a symbol. Instruction ignored.\n",
	str1, str2, ins->line);
    return NULL;
  }
}

/* ---------------------------------------------------------------------------- */
/* Function: exec_instruction							*/
/* ---------------------------------------------------------------------------- */
/* Description:									*/
/* ---------------------------------------------------------------------------- */
int exec_instruction(int pc) {
  inst_t *inst;
  int i, n, lab_pc;
  int key, found, tmp;
  int addr, ref;
  int dst, src1, src2;
  oper_t *dst_op, *src1_op, *src2_op;
  symbol_t *sym;
  symbol_t *lab;
  char buf[80];

  inst = &(program[pc]);

  if(main_debug){
    printInstruction(inst);
  }
  switch(inst->type) {
    case T_OPER:
      dst_op = checkOperandInstructionVariable(inst,1,"LHS","OPER");
      if(dst_op == NULL){
        pc++;
        break;
      }
      src1_op = getSourceOperandValue(&(inst->op2));
      src2_op = getSourceOperandValue(&(inst->op3));
      if((src1_op != NULL) && (src2_op != NULL)){
        switch(inst->operator) {
          case PLUS:
            implementBinaryPlusOperator(dst_op, src1_op, src2_op);
            break;
          case MINUS:
            implementBinaryMinusOperator(dst_op, src1_op, src2_op);
            break;
          case TIMES:
            implementBinaryTimesOperator(dst_op, src1_op, src2_op);
            break;
          case DIV:
            implementBinaryDivisionOperator(dst_op, src1_op, src2_op);
            break;
          case BIT_AND:
            implementBinaryBitwiseAndOperator(dst_op, src1_op, src2_op);
            break;
          case BIT_OR:
            implementBinaryBitwiseOrOperator(dst_op, src1_op, src2_op);
            break;
          case LOG_AND:
            implementBinaryLogicalAndOperator(dst_op, src1_op, src2_op);
            break;
          case LOG_OR:
            implementBinaryLogicalOrOperator(dst_op, src1_op, src2_op);
            break;
        }
      }
      if(isProgramCounterOperand(dst_op)){
	if(debug){
	  printf(" > DEBUG: Modifying Program Counter (oper instruction)\n");
	}
	pc = program_counter.u.ival;
	incrementNumberControlFlowInstructions();
      } else {
        pc++;
      }
      break;

    case T_UNARY:
      dst_op = checkOperandInstructionVariable(inst,1,"LHS","UNARY");
      if(dst_op == NULL){
        pc++;
        break;
      }
      src1_op = getSourceOperandValue(&(inst->op2));
      switch(inst->operator) {
        case LOG_NOT:
          implementUnaryLogicalNotOperator(dst_op, src1_op);
          break;
        case BIT_NOT:
          implementUnaryBitwiseNotOperator(dst_op, src1_op);
          break;
        case MINUS:
          implementUnaryMinusOperator(dst_op, src1_op);
          break;
      }
      pc++;
      break;

    case T_GOTO:
      pc = implementGotoInstruction(pc, inst);
      incrementNumberControlFlowInstructions();
      break;

  case T_BOOL_COND:
      pc = implementBooleanCondInstruction(pc, inst);
      incrementNumberControlFlowInstructions();
      break;

  case T_COND:
      pc = implementPredCondInstruction(pc, inst);
      incrementNumberControlFlowInstructions();
      break;

  case T_COPY:
    dst_op  = checkOperandInstructionVariable(inst,1,"LHS","COPY");
    if(dst_op == NULL){
      pc++;
      break;
    }
    src1_op = evalRHSInstructionOperand(inst, &(inst->op2), "COPY");
    if(src1_op == NULL){
      pc++;
      break;
    }
    copyOperand(dst_op,src1_op);
    if(isProgramCounterOperand(dst_op)){
      if(debug){
        printf(" > DEBUG: Modifying Program Counter (copy instruction) \n");
      }
      pc = program_counter.u.ival;
    } else {
      pc++;
    }
    break;	/* end of COPY */

  case T_PUTPAR:
      if(version == 6){
        printf(" *** Error: GETPAR/PUTPAR instruction not supported in version 6. Instruction ignored\n");
        break;
      }
    src1_op = evalRHSInstructionOperand(inst, &(inst->op1), "PUTPAR");
    if(src1_op == NULL){
      pc++;
      break;
    }
    /* --- stack2[sp2++] = memory[inst->op1] ; --- */
    dst_op = &stack2[sp2];
    copyOperand(dst_op,src1_op);
    sp2 = sp2 + 1;
    pc++;
    break;

  case T_GETPAR:
      if(version == 6){
        printf(" *** Error: GETPAR/PUTPAR instruction not supported in version 6. Instruction ignored\n");
        break;
      }
    dst_op  = checkOperandInstructionVariable(inst,1,"LHS","GETPAR");
    if(dst_op == NULL){
      pc++;
      break;
    }
    /* --- memory[inst->op1] = stack2[--sp2]; --- */
    sp2 = sp2 - 1;
    if(sp2 < 0){
      printf(" *** Error: Call Stack Underflow. GETPAR Instruction ignored\n");
      clearOperand(dst_op);
    } else {
      src1_op = &stack2[sp2];
      copyOperand(dst_op,src1_op);
    }
    pc++;
    break;

  case T_CALL:
    if(version == 6){
      printf(" *** Error: CALL/RETURN instruction not supported in version 6. Instruction ignored\n");
      break;
    }
    /* --- If there is a LHS of the CALL instruction, then it needs to be a symbol of a variable --- */
    if(inst->op1.type != O_NULL){
      if(inst->op1.type == O_STRING){
	found = 0;
  	sym = lookupItem(id_hash, &found, inst->op1.u.sval);
	if(sym == NULL){
          printf(" *** Error: Undeclared variable as LHS of CALL. Instruction ignored\n");
	  pc++;
	  break;
	}
        dst_op = &stack1[sp1];
        sp1 = sp1 + 1;
/* --- previous version 
        copyOperand(dst_op,&(inst->op1));
--- */
	/* --- push the address of the return variable rather than the symbol onto the stack. --- */
        dst_op->type = O_INTEGER;
        dst_op->u.ival = sym->address;
      } else {
        printf(" *** Error: LHS of CALL must be a variable. Instruction ignored\n");
	pc++;
	break;
      }
    }

    /* --- Now check that it is not a pre-defined library call io_proc --- */
    sym = checkTargetFunctionSymbol(inst, 2, "CALL", "FCT");
    if (sym->address < 0) {
      io_proc (inst,sym);
      pc++;
    } else {
      dst_op = &stack1[sp1];
      dst_op->type   = O_INTEGER;
      dst_op->u.ival = pc + 1;
      sp1 = sp1 + 1;
      pc = sym->address;
    }
    break;

  case T_RETURN:
    if(version == 6){
      printf(" *** Error: CALL/RETURN instruction not supported in version 6. Instruction ignored\n");
      break;
    }
    sp1 = sp1 - 1;
    if(sp1 < 0){
      printf(" *** Error: Stack Underflow in RET instruction. Instruction ignored\n");
      pc = pc + 1;
      break;
    }
    dst_op = &stack1[sp1];
    if(dst_op->type != O_INTEGER){
      printf(" *** Error: Invalid return address (not an integer). Instruction ignored\n");
      pc = pc + 1;
      break;
    }
    pc = dst_op->u.ival;
    if(inst->op1.type != O_NULL){
      sp1 = sp1 - 1; 
      dst_op = &stack1[sp1];
      src1_op = &(inst->op1); 
      if(inst->op1.type == O_STRING){
        src1_op = dereferenceStringOper(inst, src1_op);
      } else { /* assume for the time being it is integer */
      }
       
      /* now we have the address from the stack top and get a pointer to the location in memory */
      if(dst_op->type != O_INTEGER){
        printf(" *** Error: Invalid return target address (not an integer). Instruction ignored\n");
     } else {
        dst_op = &(memory[dst_op->u.ival]);
        copyOperand(dst_op,src1_op);
      }
    }
    break;

  case T_MSTORE:        
    sym = getSymbolRef(inst->op1.u.sval);
    tmp = sym->address;    
    // --- index
    if (inst->op2.type == O_STRING)     {
        sym = getSymbolRef(inst->op2.u.sval);
        tmp = tmp + memory[sym->address].u.ival;
    } else {
        tmp = tmp + inst->op2.u.ival;
    }
    
    if (inst->op3.type == O_STRING)    {
        sym = getSymbolRef(inst->op3.u.sval);
        src1_op = &(memory[sym->address]);
    } else {
        src1_op = &(inst->op3);        
    } 
    
    dst_op = &(memory[tmp]);
    copyOperand(dst_op,src1_op);            
    pc++;
    break;

  case T_MLOAD:
    sym = getSymbolRef(inst->op1.u.sval);
    tmp = sym->address;        
    // --- index
    if (inst->op2.type == O_STRING)    {
        sym = getSymbolRef(inst->op2.u.sval);
        tmp = tmp + memory[sym->address].u.ival;
    } else {
        // --- integer
        tmp = tmp + inst->op2.u.ival;
    }
        
    if (inst->op3.type == O_STRING)    {
        sym = getSymbolRef(inst->op3.u.sval);
        src1_op = &(memory[sym->address]);
    } else {
        src1_op = &(inst->op3);
    }
    
    dst_op = &(memory[tmp]);
    copyOperand(src1_op,dst_op);            
    pc++;
    break;

  case T_ADDRESSOF:
    dst_op  = checkOperandInstructionVariable(inst,1,"LHS","ADDRESSOF");
    if(dst_op == NULL){
      printf(" *** Error: Failed to eval LHS of \'addressof\' instruction in line %d; Ignored.\n",
	inst->line);
      pc++;
      break;
    }
    if(inst->op2.type == O_STRING){
      sym = getAddressableSymbolRef(inst->op2.u.sval,inst,"ADDRESSOF");
      if(sym != NULL){
        dst_op->type   = O_INTEGER;
        dst_op->u.ival = sym->address;
        if(isProgramCounterOperand(dst_op)){
	  if(debug){
            printf(" > DEBUG: Modifying Program Counter (copy instruction) \n");
	  }
          pc = program_counter.u.ival;
          incrementNumberControlFlowInstructions();
        } else {
          pc++;
        }
        break;
      }
      printf(" *** Error: Cannot take address of symbol \'%s\' in instruction in line %d\n",
	inst->op2.u.sval, inst->line);
      printf("            Must be either, a variables, a function name of a label.\n");
    }
    pc++;
    break;

  case T_DEREF:
    dst_op  = checkOperandInstructionVariable(inst,1,"LHS","DEREF");
    if(dst_op == NULL){
      printf(" *** Error: Failed to dereference LHS of PTR_DEREF instruction in line %d. Instruction ignored.\n",inst->line);
      pc++;
      break;
    }
    src1_op  = checkOperandInstructionVariable(inst,2,"RHS","DEREF");
    if(src1_op == NULL){
      printf(" *** Error: Failed to dereference RHS of PTR_DEREF instruction in line %d. Instruction ignored.\n",inst->line);
      pc++;
      break;
    }
    if(src1_op->type != O_INTEGER){
      printf(" *** Error: Invalid address in RHS of PTR_DEREF instruction in line %d. Instruction ignored.\n",inst->line);
      pc++;
      break;
    }
    ref = src1_op->u.ival;
    src1_op = &(memory[ref]);
    copyOperand(dst_op, src1_op);
    pc++;
    break; /* --- end of DEREF --- */

  case T_PTRASSIGN:
    dst_op  = checkOperandInstructionVariable(inst,1,"LHS","PTRASSIGN");
    if(dst_op == NULL){
      printf(" *** Error: Failed to dereference LHS of PTR_ASSIGN instruction in line %d. Instruction ignored.\n",inst->line);
      pc++;
      break;
    }
    if(dst_op->type != O_INTEGER){
      printf(" *** Error: Invalid address in LHS of PTR_ASSIGN instruction in line %d. Instruction ignored.\n",inst->line);
      pc++;
      break;
    }
    src1_op = &(inst->op2);
    if(src1_op->type == O_STRING){
      src1_op  = checkOperandInstructionVariable(inst,2,"RHS","PTRASSIGN");
    }
    ref = dst_op->u.ival;
    dst_op = &(memory[ref]);
    copyOperand(dst_op, src1_op);
    pc++;
    break; /* --- end of PTR_ASSIGN --- */

  case T_NOP:
    pc++;
    break;

  case T_HALT:
    pc = -1;
    break;
  }

  return(pc); 
}

/* -------------------------------------------------------------------- */
int created(char *st) {
  int key, val, found;
  char *newPtr;
  symbol_t *sym;

  found = 0;
  sym = lookupItem(id_hash, &found, st);
  if (sym == NULL) {
    val = 1;
  } else {
    val = 0;
  }
  return(val);
}

/* -------------------------------------------------------------------- */
symbol_t* getLabelRef(char *st) {
  symbol_t *sym;
  int key, found;

  found = 0;
  sym = lookupItem(label_hash, &found, st);
  if (found == 0) {
    /* --- New entry --- */
    key = label_cnt;
    /* printf(" key : %d \n",key); */
    sym = createSymbol(LAB_SYM, st, 0, 0, 0);
    labels[key] = sym;
    insertItem(label_hash,key,sym);
    label_cnt++;
  }
  return sym;
}

/* -------------------------------------------------------------------- */
symbol_t* getRegisterRef(char *st) {
  symbol_t *sym;

printf(" *** Warn: Function getRegisterRef superseeded by getRegisterSymbol... please fix code\n");

  if(st == NULL){
    return NULL;
  }
  if(isRegisterName(st)){
    return getRegisterSymbol(st);
  }
  return NULL;
}

/* -------------------------------------------------------------------- */
symbol_t* getFunctionRef(char *st) {
  symbol_t *sym;
  int key, found;

  found = 0;
  sym = lookupItem(fct_hash, &found, st);
  if (found == 0) {
    /* --- New entry --- */
    key = label_cnt;
    sym = createSymbol(FUNC_SYM, st,0,0, 0);
    labels[key] = sym;
    insertItem(fct_hash,key,sym);
    label_cnt++;
  }
  return sym;
}

/* -------------------------------------------------------------------- */
/* Function: getSymbolRef						*/
/* -------------------------------------------------------------------- */
/* Description:								*/
/*   Returns the pointer to the symbol given the symbol string.		*/
/* -------------------------------------------------------------------- */
symbol_t* getSymbolRef(char *st) {
  symbol_t *sym;
  int key, found;

  found = 0;
  /* 1. checks what kind of symbol it might be: variable vs. register */
  if(isRegisterName(st)){
    // sym = lookupItem(id_hash, &found, st);
    sym = getRegisterSymbol(st);
    if (sym == NULL) {
      fprintf(stderr," *** Error: Internal error. isRegisterSymbol and getRegisterSymbol inconsistency. (getSymbolRef)\n"); 
      return NULL;
    }
  } else {
    sym = lookupItem(id_hash, &found, st);
    if (sym == NULL) {
      /* --- New entry --- */
      key = id_cnt;
      sym = createSymbol(VAR_SYM, st, 0, 0, 0);	/* type size should go in here */
      ids[key] = sym;
      insertItem(id_hash,key,sym);
      id_cnt++;
    }
  }
  return sym;
}

/* -------------------------------------------------------------------- */

/* -------------------------------------------------------------------- */
int exec_io_proc (int address) {
  int add, ival;
  char   cval;
  double dval;
  float fval;
  oper_t *op;
  oper_t inop;

  switch (address) {
    case CREAD_ADDRESS:		/* --- cread  --- */
      if(debug){
        printf(" > DEBUG: executing \'cread\' I/O function.\n");
      }
      scanf("%c",&cval);
      inop.type = O_CHAR;
      inop.u.cval = cval;
      storeValueAsReturnValueInFrame(&inop);
      op = retrieveReturnAddressFromFrame();
      ival = op->u.ival;
      return_address.u.ival = ival;
      program_counter.u.ival = ival;
      break;

    case CWRITE_ADDRESS:	/* --- cwrite --- */
      if(debug){
        printf(" > DEBUG: executing \'cwrite\' I/O function.\n");
      }
      op = retrieveFirstArgumentFromFrame();
      cval = op->u.cval;
      if(ofp != NULL){
        fprintf(ofp,"%c",cval);
      } else {
        printf("%c",cval);
      }
      op = retrieveReturnAddressFromFrame();
      ival = op->u.ival;
      return_address.u.ival = ival;
      program_counter.u.ival = ival;
      break;

    case IREAD_ADDRESS:		/* --- iread  --- */
      if(debug){
        printf(" > DEBUG: executing \'iread\' I/O function.\n");
      }
      scanf("%d",&ival);
      inop.type = O_INTEGER;
      inop.u.ival = ival;
      storeValueAsReturnValueInFrame(&inop);
      op = retrieveReturnAddressFromFrame();
      ival = op->u.ival;
      return_address.u.ival = ival;
      program_counter.u.ival = ival;
      break;

    case IWRITE_ADDRESS:	/* --- iwrite --- */
      if(debug){
        printf(" > DEBUG: executing \'iwrite\' I/O function.\n");
      }
      op = retrieveFirstArgumentFromFrame();
      ival = op->u.ival;
      if(ofp != NULL){
        fprintf(ofp,"%d",ival);
      } else {
        printf("%d",ival);
      }
      op = retrieveReturnAddressFromFrame();
      ival = op->u.ival;
      return_address.u.ival = ival;
      program_counter.u.ival = ival;
      break;

    case FREAD_ADDRESS:		/* --- fread  --- */
      if(debug){
        printf(" > DEBUG: executing \'fread\' I/O function.\n");
      }
      scanf("%f",&fval);
      inop.type = O_DOUBLE;
      inop.u.dval = (double)fval;
      storeValueAsReturnValueInFrame(&inop);
      op = retrieveReturnAddressFromFrame();
      ival = op->u.ival;
      return_address.u.ival = ival;
      program_counter.u.ival = ival;
      break;

    case FWRITE_ADDRESS:	/* --- fwrite --- */
      if(debug){
        printf(" > DEBUG: executing \'fwrite\' I/O function.\n");
      }
      op = retrieveFirstArgumentFromFrame();
      dval = op->u.dval;
      if(ofp != NULL){
        fprintf(ofp,"%f",dval);
      } else {
        printf("%f",dval);
      }
      op = retrieveReturnAddressFromFrame();
      ival = op->u.ival;
      return_address.u.ival = ival;
      program_counter.u.ival = ival;
      break;

    default:
      printf (" *** Error: Unknown I/O function. Trap to monitor... Exit(1).\n");
      return 1;
      break;
  }
  return 0;
}

void io_proc (inst_t *ins, symbol_t *sym) {
  int addr, ival;
  float fval;
  oper_t *myref, *dst_op, *src1_op;

  switch (sym->address) {
    case CREAD_ADDRESS:		/* --- cread --- */
      sp1 = sp1 - 1;
      if(sp1 < 0){
	printf(" *** Error: Stack underflow in cread I/O function invocation. Ignored.\n");
	break;
      }
      myref = &stack1[sp1];
      if(myref->type == O_INTEGER){
	addr = myref->u.ival;
      } else {
	printf(" *** Error: Cannot dereference address taken from the stack top. Instruction ignored\n");
        break;
      }
      dst_op = &(memory[addr]);
      if(dst_op == NULL){
	printf(" *** Error: Cannot dereference variable %s in cread I/O function invocation.\n",
	  myref->u.sval);
      } else {
        dst_op->u.cval = (char) getc(stdin); 
	dst_op->type = O_CHAR;
      }
      break;      

    case CWRITE_ADDRESS:		/* --- cwrite --- */
      sp2 = sp2 - 1;
      if(sp1 < 0){
	break;
      }
      myref = &stack2[sp2];
      if(myref->type != O_CHAR){
        printf (" *** Error: Argument of cwrite function in not a character. Skipped\n");
      } else {
	if(ofp != NULL){
          fprintf (ofp,"%c", (char) myref->u.cval);
	} else {
          printf ("%c", (char) myref->u.cval);
	}
      }
      break;      

    case IREAD_ADDRESS:		/* --- iread --- */
      sp1 = sp1 - 1;
      if(sp1 < 0){
	printf(" *** Error: Stack underflow in iread I/O function invocation. Ignored.\n");
	break;
      }
      myref = &stack1[sp1];
      if(myref->type == O_INTEGER){
	addr = myref->u.ival;
      } else {
	printf(" *** Error: Cannot dereference address taken from the stack top. Instruction ignored\n");
        break;
      }
      dst_op = &(memory[addr]);
      if(dst_op == NULL){
	printf(" *** Error: Cannot dereference variable %s in iread I/O function invocation.\n",
	  myref->u.sval);
      } else {
        scanf ("%d", &ival);
        dst_op->u.ival = ival;
	dst_op->type = O_INTEGER;
      }
      break;      

    case IWRITE_ADDRESS:		/* --- iwrite --- */
      sp2 = sp2 - 1;
      myref = &stack2[sp2];
      if(myref->type != O_INTEGER){
        printf (" *** Error: Argument of iwrite function in not an integer. Skipped\n");
      } else {
        if(ofp != NULL){
          fprintf(ofp,"%d", (int) myref->u.ival);
        } else {
          printf ("%d", (int) myref->u.ival);
	}
      }
      break;      

    case FREAD_ADDRESS:		/* --- fread --- */
      sp1 = sp1 - 1;
      if(sp1 < 0){
	printf(" *** Error: Stack underflow in fread I/O function invocation. Ignored.\n");
	break;
      }
      myref = &stack1[sp1];
      if(myref->type == O_INTEGER){
	addr = myref->u.ival;
      } else {
	printf(" *** Error: Cannot dereference address taken from the stack top. Instruction ignored\n");
        break;
      }

      dst_op = &(memory[addr]);
      if(dst_op == NULL){
	printf(" *** Error: Cannot dereference variable %s in fread I/O function invocation.\n",
	  myref->u.sval);
      } else {
        scanf ("%g", &fval);
        dst_op->u.dval = (double)fval;
	dst_op->type = O_DOUBLE;
      }
      break;      

    case FWRITE_ADDRESS:		/* fwrite */
      sp2 = sp2 - 1;
      myref = &stack2[sp2];
      if(myref->type != O_DOUBLE){
        printf (" *** Error: Argument of fwrite function in not a double. Skipped\n");
      } else {
        if(ofp != NULL){
          fprintf(ofp,"%f", (double) myref->u.dval);
        } else {
          printf ("%f", (double) myref->u.dval);
	}
      }
      break;      

    default:
      printf (" *** Error: Unknown I/O function. Trap to monitor... exit(1).\n");
      break;      
    ;
  }
}

/* -------------------------------------------------------------------- */
/*

$Log: main.c,v $

*/
