%{
/* **************************************************************************** */
/*                                                                              */
/*  Copyright (c) 2011 University of Southern California (USC)                  */ 
/*                                                                              */
/*  All rights reserved.                                                        */
/*                                                                              */
/*  Permission is given to use, copy, and modify this software for any          */
/*  non-commercial purpose as long as this copyright notice is not removed      */
/*  All other uses, including redistribution in whole or in part, are           */
/*  forbidden without prior written permission.                                 */
/*                                                                              */
/*  This software is provided with absolutely no warranty and no support.       */
/*                                                                              */
/*  For further information contact: pedro@isi.edu                              */
/*                                                                              */
/* **************************************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "y.tab.h"
#include "structs.h"

static int comment(void);

#ifdef DEBUG
#define ECHO printf("%s",yytext);
#else
#define ECHO
#endif

extern int file_line;
extern int lineno;
extern int program_line;

extern int lex_debug;

void setlval();
void setsval();
char* strsav(char *s);
%}
B                       [0-1]
O                       [0-7]
D                       [0-9]
L                       [a-zA-Z_@]
H                       [a-fA-F0-9]
BS                      (b|B)
OS                      (o|O)
DS                      (d|D)
HS                      (h|H)
Space			[ \t]
%%
^#		{ if(lex_debug) printf(" > LEX removing comment in line %d\n",lineno); comment(); }
"\$ra"		{ if(lex_debug) printf(" > LEX returning return address register\n");
		 // yylval.i = RA_REG;
	 	 //  return (REG_ID); 
	         setlval();
	 	 return (ID); 
		}
"\$sp"		{ if(lex_debug) printf(" > LEX returning stack pointer register\n");
		 // yylval.i = SP_REG;
	 	 //  return (REG_ID); 
	         setlval();
	 	 return (ID); 
		}
"\$pc"		{ if(lex_debug) printf(" > LEX returning program counter register\n"); 
		 //  yylval.i = PC_REG;
		 //  return (REG_ID); 
	         setlval();
	 	 return (ID); 
		}
"\$fp"		{ if(lex_debug) printf(" > LEX returning frame pointer register\n"); 
		  // yylval.i = FP_REG; 
		  // return (REG_ID); 
	         setlval();
	 	 return (ID); 
		}
","             { if(lex_debug) printf(" > LEX returning COMMA\n"); 	return(COMMA);  }
":"             { if(lex_debug) printf(" > LEX returning COLON\n"); 	return(COLLON); }
"="             { if(lex_debug) printf(" > LEX returning ATRIB\n"); 	return(ATRIB);  }
"+"             { if(lex_debug) printf(" > LEX returning PLUS\n");  	return(PLUS);   }
"-"             { if(lex_debug) printf(" > LEX returning MINUS\n"); 	return(MINUS);  }
"*"             { if(lex_debug) printf(" > LEX returning TIMES\n"); 	return(TIMES);  }
"/"             { if(lex_debug) printf(" > LEX returning '/'\n");   	return(DIV);   	}
"&&"            { if(lex_debug) printf(" > LEX returning '&'\n");   	return(LOG_AND);}
"||"            { if(lex_debug) printf(" > LEX returning OR\n");    	return(LOG_OR); }
"!"             { if(lex_debug) printf(" > LEX returning NOT\n"); 	return(LOG_NOT);}
"&"             { if(lex_debug) printf(" > LEX returning '&'\n");   	return(BIT_AND);}
"|"             { if(lex_debug) printf(" > LEX returning OR\n");    	return(BIT_OR); }
"~"             { if(lex_debug) printf(" > LEX returning NOT\n"); 	return(BIT_NOT);}
"\["            { if(lex_debug) printf(" > LEX returning LBR\n"); 	return(LBR); 	}
"\]"            { if(lex_debug) printf(" > LEX returning RBR\n"); 	return(RBR); 	}
"\("            { if(lex_debug) printf(" > LEX returning LPAREN\n"); 	return(LPAREN); }
"\)"            { if(lex_debug) printf(" > LEX returning RPAREN\n"); 	return(RPAREN); }
">"             { if(lex_debug) printf(" > LEX returning GT\n"); 	return(GT); 	}
"<"             { if(lex_debug) printf(" > LEX returning LT\n"); 	return(LT); 	}
">="            { if(lex_debug) printf(" > LEX returning GE\n"); 	return(GE); 	}
"<="            { if(lex_debug) printf(" > LEX returning LE\n"); 	return(LE); 	}
"=="            { if(lex_debug) printf(" > LEX returning EQ\n"); 	return(EQ); 	}
"!="            { if(lex_debug) printf(" > LEX returning NEQ\n"); 	return(NEQ); 	}
"goto"          { if(lex_debug) printf(" > LEX returning GOTO\n"); 	return(GOTO); 	}
"jump"          { if(lex_debug) printf(" > LEX returning JUMP\n"); 	return(JUMP); 	}
"if"            { if(lex_debug) printf(" > LEX returning IF\n"); 	return(IF); 	}
"halt"          { if(lex_debug) printf(" > LEX returning HALT\n"); 	return(HALT); 	}
"putparam"      { if(lex_debug) printf(" > LEX returning PUTPAR\n"); 	return(PUTPAR); }
"getparam"      { if(lex_debug) printf(" > LEX returning GETPAR\n"); 	return(GETPAR); }
"return"        { if(lex_debug) printf(" > LEX returning RETURN\n"); 	return(RETURN); }
"call"          { if(lex_debug) printf(" > LEX returning CALL\n"); 	return(CALL); 	}
"phi"           { if(lex_debug) printf(" > LEX returning PHI\n"); 	return(PHI); 	}
"nop"           { if(lex_debug) printf(" > LEX returning NOP\n"); 	return(NOP); 	}
".section"	{ if(lex_debug) printf(" > LEX returning directive (%s)\n",yytext); 	return(SEC_DIR); 	}
".skip"		{ if(lex_debug) printf(" > LEX returning directive (%s)\n",yytext); 	return(SKIP_DIR); 	}
".data"		{ if(lex_debug) printf(" > LEX returning directive (%s)\n",yytext); 	return(DATA_DIR); 	}
".text"		{ if(lex_debug) printf(" > LEX returning directive (%s)\n",yytext); 	return(TEXT_DIR); 	}
".word"		{ if(lex_debug) printf(" > LEX returning directive (%s)\n",yytext); 	return(WORD_DIR); 	}
".byte"		{ if(lex_debug) printf(" > LEX returning directive (%s)\n",yytext); 	return(BYTE_DIR); 	}
".ascii"	{ if(lex_debug) printf(" > LEX returning directive (%s)\n",yytext); 	return(ASCII_DIR); 	}
".global"	{ if(lex_debug) printf(" > LEX returning directive (%s)\n",yytext); 	return(GLOBAL_DIR); 	}
".end"		{ if(lex_debug) printf(" > LEX returning directive (%s)\n",yytext); 	return(END_DIR); 	}
\$r({D}|{D}{D}) { if(lex_debug) printf(" > LEX returning register (%s)\n",yytext); 	
		// yylval.i = atoi(&yytext[2]);
		//return(REG_ID); 
		  setlval(); 
		  return(ID); 
		}
{L}({L}|{D})*   { if(lex_debug) printf(" > LEX returning ID\n"); 
		  setlval(); 
		  return(ID); 
		}
[+-]?{D}{D}*   	{ 
		  if(lex_debug) 
		    printf(" > LEX returning IVAL\n"); 
		  yylval.i = atoi(yytext); 
		  return(IVAL); 
		} 
[+-]?{D}{D}*(\.{D}*)?   { /* setlval(); */ 
		  if(lex_debug) printf(" > LEX returning DVAL\n");
		    yylval.d = atof(yytext); 
		  return(DVAL); 
		} 
[\n]		{ 
		  if(lex_debug) 
		    printf(" > LEX returning NL\n");
		  file_line++; 
		  lineno++;
		  return (NL);  
		}
\"(({L}|{D}|{Space})*)\"	{
		  setsval(); 
		  if(lex_debug) 
		    printf(" > LEX returning SVAL (%s)\n",&(yytext[1]));
		  return (SVAL);
		}
\'[^.]\'	{
		  if(lex_debug) 
		    printf(" > LEX returning CVAL\n");
		  yylval.c = (yytext[1]); 
		  return (CVAL);  
		}
.		{
		  if(lex_debug) 
		  printf(" > LEX: matching unexpected (possibly empty) text (%s)\n",yytext);
		}
%%

/* --------------------------------------------------------------------------- */

int comment(){
  int ch;
  if(lex_debug) {
    printf(" > LEX skipping comment.\n");
  }
  for(;;) {
    ch = input();
    if((ch == '\n') || (ch == '\0')){
      return 0;
    }
  }
  return 1;
}

/* --------------------------------------------------------------------------- */

void setlval() {
  yylval.s = (char *) malloc(strlen(yytext)+1);
  if (yylval.s == NULL)  {
    printf(" *** Error: out of memory inside setlval. (malloc) Exiting...\n");
    exit(-1);
  }
  strcpy(yylval.s, yytext);
}

/* --------------------------------------------------------------------------- */

void setsval() {
  yylval.s = (char *) malloc(strlen(yytext)-1);
  if (yylval.s == NULL)  {
    printf(" *** Error: out of memory inside setsval. (malloc) Exiting...\n");
    exit(-1);
  }
  yytext[strlen(yytext)-1] = '\0';
  strcpy(yylval.s, &(yytext[1]));
}

/* --------------------------------------------------------------------------- */

char* strsav(char *s) {
  char *t;
  t = (char*)malloc(strlen(s)+1);
  if(t == NULL){
    printf(" *** Error: out of memory inside strsav. (malloc) Exiting...\n");
  }
  strcpy(t,s);
  return t;
}

/* --------------------------------------------------------------------------- */

