/* **************************************************************************** */
/*                                                                              */
/*  Copyright (c) 2011 University of Southern California (USC)                  */ 
/*                                                                              */
/*  All rights reserved.                                                        */
/*                                                                              */
/*  Permission is given to use, copy, and modify this software for any          */
/*  non-commercial purpose as long as this copyright notice is not removed      */
/*  All other uses, including redistribution in whole or in part, are           */
/*  forbidden without prior written permission.                                 */
/*                                                                              */
/*  This software is provided with absolutely no warranty and no support.       */
/*                                                                              */
/*  For further information contact: pedro@isi.edu                              */
/*                                                                              */
/* **************************************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "structs.h"

extern char * strsav(char *s);

extern oper_t   program_counter;
extern symbol_t program_counter_sym;
extern oper_t   stack_pointer;
extern symbol_t stack_pointer_sym;
extern oper_t   frame_pointer;
extern symbol_t frame_pointer_sym;
extern oper_t   return_address;
extern symbol_t return_address_sym;
extern oper_t   reg[NUMBER_REGISTERS];
extern symbol_t reg_sym[NUMBER_REGISTERS];

/* ----------------------------------------------------------------------- */
int isNullOperand(oper_t* op){
  if(op == NULL){
    return 0;
  }
  if(op->type == O_NULL){
    return 1;
  }
  return 0;
}

/* ----------------------------------------------------------------------- */
int isDoubleOperand(oper_t* op){
  if(op == NULL){
    return 0;
  }
  if(op->type == O_DOUBLE){
    return 1;
  }
  return 0;
}

/* ----------------------------------------------------------------------- */
int isStringOperand(oper_t* op){
  if(op == NULL){
    return 0;
  }
  if(op->type == O_STRING){
    return 1;
  }
  return 0;
}

/* ----------------------------------------------------------------------- */
int isCharOperand(oper_t* op){
  if(op == NULL){
    return 0;
  }
  if(op->type == O_CHAR){
    return 1;
  }
  return 0;
}

/* ----------------------------------------------------------------------- */
int isProgramCounterOperand(oper_t* op){
  if(op == NULL){
    return 0;
  }
  if(op == &program_counter){
    return 1;
  }
  return 0;
}

/* ----------------------------------------------------------------------- */
void printOperand(oper_t* op){
  if(op == NULL){
    printf(" null oper");
  } else {
    if(op->type == O_INTEGER){
      printf(" op(i): %d",op->u.ival);
      return;
    }
    if(op->type == O_DOUBLE){
      printf(" op(d): %f",op->u.dval);
      return;
    }
    if(op->type == O_NULL){
      printf(" op(n): nil");
      return;
    }
    if(op->type == O_STRING){
      printf(" op(s): %s",op->u.sval);
      return;
    }
    if(op->type == O_CHAR){
      printf(" op(c): %c",op->u.cval);
      return;
    }
    return;
  }
}

/* ----------------------------------------------------------------------- */
void setOperandInteger(oper_t *dst, int v){
  dst->type = O_INTEGER;
  dst->u.ival = v;
}

/* ----------------------------------------------------------------------- */
int getIntegerOperandValue(oper_t *src){
  int res;
  if(src == NULL){
    fprintf(stderr, " *** Error: Nil Operand Reference. -1 value returned. (getIntegerOperandValue)\n");
    return -1;
  }

  if(src->type != O_INTEGER){
    fprintf(stderr, " *** Error: Operand is not an integer. -1 value returned. (getIntegerOperandValue)\n");
    return -1;
  }

  res = src->u.ival;
  return res;
}

/* ----------------------------------------------------------------------- */
void copyOperand(oper_t* dst, oper_t* src){
  dst->type = src->type;
  if(src->type == O_INTEGER){
    dst->u.ival = src->u.ival;
    return;
  }
  if(src->type == O_DOUBLE){
    dst->u.dval = src->u.dval;
    return;
  }
  if(src->type == O_STRING){
    dst->u.sval = strsav(src->u.sval);
    return;
  }
  if(src->type == O_CHAR){
    dst->u.cval = src->u.cval;
    return;
  }
}

/* ----------------------------------------------------------------------- */
oper_t *createStringValOperand(char *str){
  oper_t *res;

  res = (oper_t*)malloc(sizeof(oper_t));
  res->type = O_STRING;
  res->u.sval = str;
  return res;
}

/* ----------------------------------------------------------------------- */
oper_t *createIntegerValOperand(int v){
  oper_t *res;

  res = (oper_t*)malloc(sizeof(oper_t));
  res->type = O_INTEGER;
  res->u.ival = v;
  return res;
}

/* ----------------------------------------------------------------------- */
oper_t *createCharValOperand(char v){
  oper_t *res;

  res = (oper_t*)malloc(sizeof(oper_t));
  res->type = O_CHAR;
  res->u.cval = v;
  return res;
}

/* ----------------------------------------------------------------------- */
oper_t *createDoubleValOperand(double v){
  oper_t *res;

  res = (oper_t*)malloc(sizeof(oper_t));
  res->type = O_DOUBLE;
  res->u.dval = v;
  return res;
}

/* ----------------------------------------------------------------------- */
void clearOperand(oper_t *op){
  if(op != NULL){
    op->type = O_NULL;
  }
}

/* ----------------------------------------------------------------------- */
void destroyOperand(oper_t *op){
  if(op != NULL){
    if(op->type == O_INTEGER)
      op->u.ival = 0;
    if(op->type == O_DOUBLE)
      op->u.dval = 0;
    if(op->type == O_STRING){
      if(op->u.sval != NULL){
        free(op->u.sval);
        op->u.sval = NULL;
      }
    }
    if(op->type == O_CHAR)
      op->u.cval = '\0';
    op->type = O_NULL;
    free(op);
  }
}

/* -------------------------------------------------------------------- */
/* instruction data structure functions					*/
/* -------------------------------------------------------------------- */
void printInstructionType(int t){

  switch(t){
    case T_VOID: 	printf(" VOID "); 	break;
    case T_OPER:	printf(" OPER "); 	break;
    case T_UNARY:	printf(" UNARY ");	break;
    case T_COPY: 	printf(" COPY "); 	break;
    case T_MSTORE: 	printf(" MSTORE "); 	break;
    case T_MLOAD: 	printf(" MLOAD "); 	break;
    case T_ADDRESSOF: 	printf(" ADDRESSOF "); 	break;
    case T_DEREF: 	printf(" DREF "); 	break;
    case T_PTRASSIGN: 	printf(" PTRASSIGN "); 	break;
    case T_COND: 	printf(" COND "); 	break;
    case T_BOOL_COND: 	printf(" BOOL_COND "); 	break;
    case T_GOTO: 	printf(" GOTO "); 	break;
    case T_HALT: 	printf(" HALT "); 	break;
    case T_PUTPAR: 	printf(" PUTPAR "); 	break;
    case T_GETPAR: 	printf(" GETPAR "); 	break;
    case T_CALL: 	printf(" CALL "); 	break;
    case T_RETURN: 	printf(" RETURN "); 	break;
    case T_JUMP: 	printf(" JUMP "); 	break;
    case T_NOP: 	printf(" NOP "); 	break;
    default: 		printf(" undef "); 	break;
  }
}

/* -------------------------------------------------------------------- */
void printInstructionOperator(int op){
  switch(op){
    case PLUS: 		printf(" PLUS "); 	break;
    case MINUS: 	printf(" MINUS "); 	break;
    case TIMES: 	printf(" TIMES "); 	break;
    case DIV: 		printf(" DIV "); 	break;
    case BIT_AND: 	printf(" BIT AND "); 	break;
    case BIT_OR: 	printf(" BIT OR "); 	break;
    case LOG_AND: 	printf(" LOG AND "); 	break;
    case LOG_OR: 	printf(" LOG OR "); 	break;
    case PHI: 		printf(" PHI "); 	break;
    case LT: 		printf(" LT "); 	break;
    case GT: 		printf(" GT "); 	break;
    case LE: 		printf(" LE "); 	break;
    case GE: 		printf(" GE "); 	break;
    case EQ: 		printf(" EQ "); 	break;
    case NEQ: 		printf(" NEQ "); 	break;
    case BIT_NOT: 	printf(" BIT NOT "); 	break;
    case LOG_NOT: 	printf(" LOG NOT "); 	break;
    default: 		printf(" <nop> "); 	break;
  }
}

/* -------------------------------------------------------------------- */
void printInstruction(inst_t *ip){
  if(ip != NULL){
    printf(" type: "); printInstructionType(ip->type); 		printf("\n");
    printf(" op1: ");  printOperand(&(ip->op1)); 		printf("\n");
    printf(" op2: ");  printOperand(&(ip->op2)); 		printf("\n");
    printf(" op3: ");  printOperand(&(ip->op3)); 		printf("\n");
    printf(" op: ");   printInstructionOperator(ip->operator); 	printf("\n");
  }
}

/* -------------------------------------------------------------------- */
void clearInstruction(inst_t *ip){
  if(ip != NULL){
    ip->type = T_VOID;
    ip->op1.type = O_NULL;
    ip->op2.type = O_NULL;
    ip->op3.type = O_NULL;
    ip->operator = O_NULL;
  }
}

/* -------------------------------------------------------------------- */
/* symbol data structure functions					*/
/* -------------------------------------------------------------------- */

void initSymbol(symbol_t *sym, int t, char *s, int n, int sz, int att){
  if(sym == NULL){
    fprintf(stderr," *** Error: Invalid nil symbol. (initSymbol)\n");
  } else {
    setSymbolType(sym, t);
    setSymbolTag(sym, s);
    setSymbolAddress(sym, n);
    setSymbolSize(sym, sz);
    setSymbolAttribute(sym, att);
  }
}

symbol_t *createSymbol(int t, char *s, int n, int sz, int att){
  symbol_t *res;

  res = (symbol_t*)malloc(sizeof(symbol_t));
  if(res == NULL){
    fprintf(stderr," *** Error: Cannot allocate storage for symbol. Nil returned (createSymbol)\n");
    return NULL;
  }
  initSymbol(res,t,s,n,sz,att);
  return res;
}

void destroySymbol(symbol_t *s){
  if(s == NULL){
    s->type 	= NULL_SYM;
    s->tag[0] 	= '\0';
    s->address 	= 0;
    s->size 	= 0;
    s->attrib 	= 0;
    if(s->initvalue != NULL){
      destroyOperand(s->initvalue);
    }
    s->initvalue = NULL;
    free(s);
  }
}

void printSymbol(symbol_t *s){
  if(s == NULL){
    printf(" (nil symbol)");
  } else {
    printf(" (%d:%s:%d:%d:%d",s->type,s->tag, s->address, s->size,s->attrib);
    if(s->initvalue == NULL){
      printf("i:<nil>)"); 
    } else {
      printf("i:"); 
      printOperand(s->initvalue);
      printf(")");
    }
  }
}

void setSymbolType(symbol_t *s, int t){
  if(s != NULL){
    s->type = t;
  }
}

void setSymbolAttribute(symbol_t *s, int att){
  if(s != NULL){
    s->attrib = att;
  }
}

void setSymbolTag(symbol_t *s, char *t){
  if(s != NULL){
    strncpy(s->tag, t, MAX_SYMBOL_TAG_SIZE-2);
    s->tag[MAX_SYMBOL_TAG_SIZE-1] = '\0';
  }
}

void setSymbolAddress(symbol_t *s, int addr){
  if(s != NULL){
    s->address = addr;
  }
}

void setSymbolSize(symbol_t *s, int v){
  if(s != NULL){
    s->size = v;
  }
}

void setSymbolInitValue(symbol_t *s, oper_t* v){
  if(s != NULL){
    s->initvalue = v;
  }
}


/* -------------------------------------------------------------------- */
/* Functions for Variable Symbols					*/
/* -------------------------------------------------------------------- */
int isVariableSymbol(symbol_t *s){
  if(s == NULL){
    return 0;
  }
  if(s->type == VAR_SYM)
    return 1;
  return 0;
}

/* -------------------------------------------------------------------- */
/* Functions for Procedure/Function Symbols				*/
/* -------------------------------------------------------------------- */
int isFunctionSymbol(symbol_t *s){
  if(s == NULL){
    return 0;
  }
  if(s->type == FUNC_SYM)
    return 1;
  return 0;
}

int isProcecureSymbol(symbol_t *s){
  if(s == NULL){
    return 0;
  }
  if(s->type == PROC_SYM)
    return 1;
  return 0;
}

/* -------------------------------------------------------------------- */
/* Functions for Register Symbols					*/
/* -------------------------------------------------------------------- */
int isRegisterName(char *s){
  int n;

  if(s == NULL){
    return 0;
  }
  if(strcmp(s,"$ra") == 0){
    return 1;
  }
  if(strcmp(s,"$pc") == 0){
    return 1;
  }
  if(strcmp(s,"$sp") == 0){
    return 1;
  }
  if(strcmp(s,"$fp") == 0){
    return 1;
  }
  if((s[0] == '$') && (s[1] == 'r')){
    return 1;
  }
  return 0;
}

int isRegisterSymbol(symbol_t *s){
  if(s == NULL){
    return 0;
  }
  if(s->type == REG_SYM)
    return 1;
  return 0;
}

oper_t* getRegisterSymbolOperand(char *s){
  int n;

  if(s == NULL){
    return NULL;
  }
  if(strcmp(s,"$pc") == 0){
    return &(program_counter);
  }
  if(strcmp(s,"$ra") == 0){
    return &(return_address);
  }
  if(strcmp(s,"$sp") == 0){
    return &(stack_pointer);
  }
  if(strcmp(s,"$fp") == 0){
    return &(frame_pointer);
  }
  if((s[0] == '$') && (s[1] == 'r')){
    n = atoi(&(s[2]));
    if((n >= 0) && (n <= 31)){
      return &(reg[n]);
    }
  }
  return NULL;
}

symbol_t* getRegisterSymbol(char *s){
  int n;

  if(s == NULL){
    return NULL;
  }
  if(strcmp(s,"$pc") == 0){
    return &(program_counter_sym);
  }
  if(strcmp(s,"$ra") == 0){
    return &(return_address_sym);
  }
  if(strcmp(s,"$sp") == 0){
    return &(stack_pointer_sym);
  }
  if(strcmp(s,"$fp") == 0){
    return &(frame_pointer_sym);
  }
  if((s[0] == '$') && (s[1] == 'r')){
    n = atoi(&(s[2]));
    if((n >= 0) && (n <= 31)){
      return &(reg_sym[n]);
    }
  }
  return NULL;
}
